import subprocess
import threading
import time


class LazyProcess:
    """Abstraction describing a command line launching a service - probably
    as needed as functionality is accessed in Galaxy.
    """

    def __init__(self, command_and_args):
        self.command_and_args = command_and_args
        self.thread_lock = threading.Lock()
        self.allow_process_request = True
        self.process = None

    def start_process(self):
        with self.thread_lock:
            if self.allow_process_request:
                self.allow_process_request = False
                t = threading.Thread(target=self.__start)
                t.daemon = True
                t.start()

    def __start(self):
        with self.thread_lock:
            self.process = subprocess.Popen(self.command_and_args, close_fds=True)

    def shutdown(self):
        with self.thread_lock:
            self.allow_process_request = False
        if self.running:
            assert self.process  # tell type checker it can not be None if self.running
            self.process.terminate()
            time.sleep(0.01)
            if self.running:
                self.process.kill()

    @property
    def running(self):
        return self.process and not self.process.poll()


class NoOpLazyProcess:
    """LazyProcess abstraction meant to describe potentially optional
    services, in those cases where one is not configured or valid, this
    class can be used in place of LazyProcess.
    """

    def start_process(self):
        return

    def shutdown(self):
        return

    @property
    def running(self):
        return False



import re
from typing import (
    Dict,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

KeyedQueryT = Tuple[str, str]
ParseFilterResultT = Tuple[Optional[List["FilteredTerm"]], Optional[str]]
QUOTE_PATTERN = re.compile(r"\'(.*?)\'")


def parse_filters(search_term: str, filters: Optional[Dict[str, str]] = None) -> ParseFilterResultT:
    """Support github-like filters for narrowing the results.

    Order of chunks does not matter, only recognized filter names are allowed.

    :param search_term: the original search str from user input

    :returns allow_query: whoosh Query object used for filtering
        results of searching in index
    :returns search_term_without_filters: str that represents user's
        search phrase without the filters
    """
    return parse_filters_structured(search_term, filters, preserve_quotes=False).simple_result


def parse_filters_structured(
    search_term: str,
    filters: Optional[Dict[str, str]] = None,
    preserve_quotes: bool = True,
) -> "ParsedSearch":
    search_space = search_term.replace('"', "'")
    filters = filters or {}
    filter_keys = "|".join(list(filters.keys()))
    pattern = rf"({filter_keys}):(?:\s+)?([\w-]+|\'.*?\')"
    reserved = re.compile(pattern)
    parsed_search = ParsedSearch()
    while True:
        match = reserved.search(search_space)
        if match is None:
            match = QUOTE_PATTERN.search(search_space)
            if match is None:
                parsed_search.add_unfiltered_text_terms(search_space)
                break
            group = match.groups()[0].strip()
            parsed_search.add_unfiltered_text_terms(search_space[0 : match.start()])
            parsed_search.add_unfiltered_text(group, True)
        else:
            first_group = match.groups()[0]
            if first_group in filters:
                filter_as = filters[first_group]
                group = match.groups()[1].strip()
                quoted = preserve_quotes and group.startswith("'")
                parsed_search.add_keyed_term(filter_as, group.replace("'", ""), quoted)
            parsed_search.add_unfiltered_text_terms(search_space[0 : match.start()])
        search_space = search_space[match.end() :]
    return parsed_search


class RawTextTerm(NamedTuple):
    text: str
    quoted: bool


class FilteredTerm(NamedTuple):
    filter: str
    text: str
    quoted: bool


TermT = Union[RawTextTerm, FilteredTerm]


class ParsedSearch:
    terms: List[TermT]
    text_terms: List[RawTextTerm]
    filter_terms: List[FilteredTerm]

    def __init__(self):
        self.terms = []
        self.text_terms = []
        self.filter_terms = []

    def add_unfiltered_text_terms(self, text: str):
        for part in text.split():
            self.add_unfiltered_text(part, False)

    def add_unfiltered_text(self, text: str, quoted: bool = False):
        text = text.strip()
        if not text:
            return
        term = RawTextTerm(text.strip(), quoted)
        self.terms.append(term)
        self.text_terms.append(term)

    def add_keyed_term(self, key: str, text: str, quoted: bool):
        term = FilteredTerm(key, text, quoted)
        self.terms.append(term)
        self.filter_terms.append(term)

    @property
    def simple_result(self) -> ParseFilterResultT:
        return None if len(self.filter_terms) == 0 else self.filter_terms, " ".join([t.text for t in self.text_terms])


__all__ = (
    "parse_filters",
    "parse_filters_structured",
)



import time
from os import PathLike
from typing import (
    List,
    Optional,
    Tuple,
    Union,
)

import requests

from galaxy import exceptions
from galaxy.files import (
    ConfiguredFileSources,
    FileSourceDictifiable,
)
from galaxy.files.sources import FilesSourceOptions
from galaxy.files.sources.http import HTTPFilesSourceProperties
from galaxy.files.uris import stream_url_to_file
from galaxy.util import DEFAULT_SOCKET_TIMEOUT
from galaxy.util.config_parsers import IpAllowedListEntryT

TargetPathT = Union[str, PathLike]


def _not_implemented(drs_uri: str, desc: str) -> NotImplementedError:
    missing_client_func = f"Galaxy client cannot currently fetch URIs {desc}."
    header = f"Missing client functionaltiy required to fetch DRS URI {drs_uri}."
    rest_of_message = """Currently Galaxy client only works with HTTP/HTTPS targets but extensions for
    other types would be gladly welcomed by the Galaxy team. Please
    report use cases not covered by this function to our issue tracker
    at https://github.com/galaxyproject/galaxy/issues/new.
    """
    return NotImplementedError(f"{header} {missing_client_func} {rest_of_message}")


class RetryOptions:
    retry_times: int = 5
    override_retry_after: Optional[float] = None


def retry_and_get(get_url: str, retry_options: RetryOptions, headers: Optional[dict] = None) -> requests.Response:
    response = requests.get(get_url, timeout=DEFAULT_SOCKET_TIMEOUT, headers=headers)
    response.raise_for_status()
    if response.status_code == 202:
        if retry_options.retry_times == 0:
            raise Exception("Timeout waiting on DRS URI response")
        if "Retry-After" not in response.headers:
            raise ValueError("Remote DRS URI response invalid (Retry-After must be specified for 202 statuses)")
        retry_after = retry_options.override_retry_after or float(response.headers["Retry-After"])
        time.sleep(retry_after)
        retry_options.retry_times -= 1
        return retry_and_get(get_url, retry_options)
    else:
        return response


def _get_access_info(obj_url: str, access_method: dict, headers=None) -> Tuple[str, dict]:
    try:
        access_url = access_method["access_url"]
    except KeyError:
        access_id = access_method["access_id"]
        access_get_url = f"{obj_url}/access/{access_id}"
        access_response = requests.get(access_get_url, timeout=DEFAULT_SOCKET_TIMEOUT, headers=headers)
        access_response.raise_for_status()
        access_response_object = access_response.json()
        access_url = access_response_object

    url = access_url["url"]
    headers_list = access_url.get("headers") or []
    headers_as_dict = {}
    for header_str in headers_list:
        key, value = header_str.split(": ", 1)
        headers_as_dict[key] = value

    return url, headers_as_dict


def fetch_drs_to_file(
    drs_uri: str,
    target_path: TargetPathT,
    user_context: FileSourceDictifiable,
    force_http=False,
    retry_options: Optional[RetryOptions] = None,
    headers: Optional[dict] = None,
    fetch_url_allowlist: Optional[List[IpAllowedListEntryT]] = None,
):
    """Fetch contents of drs:// URI to a target path."""
    if not drs_uri.startswith("drs://"):
        raise ValueError(f"Unknown scheme for drs_uri {drs_uri}")
    rest_of_drs_uri = drs_uri[len("drs://") :]
    if "/" not in rest_of_drs_uri:
        # DRS URI uses compact identifiers, not yet implemented.
        # https://ga4gh.github.io/data-repository-service-schemas/preview/release/drs-1.2.0/docs/more-background-on-compact-identifiers.html
        raise _not_implemented(drs_uri, "that use compact identifiers")
    netspec, object_id = rest_of_drs_uri.split("/", 1)
    scheme = "https"
    if force_http:
        scheme = "http"
    get_url = f"{scheme}://{netspec}/ga4gh/drs/v1/objects/{object_id}"
    response = retry_and_get(get_url, retry_options or RetryOptions(), headers=headers)
    response.raise_for_status()
    response_object = response.json()
    access_methods = response_object.get("access_methods", [])
    if len(access_methods) == 0:
        raise ValueError(f"No access methods found in DRS response for {drs_uri}")

    downloaded = False
    for access_method in access_methods:
        access_url, access_headers = _get_access_info(get_url, access_method, headers=headers)
        opts = FilesSourceOptions()
        if access_method["type"] == "https":
            extra_props: HTTPFilesSourceProperties = {
                "http_headers": access_headers or {},
                "fetch_url_allowlist": fetch_url_allowlist or [],
            }
            opts.extra_props = extra_props
        else:
            opts.extra_props = {}
        try:
            file_sources = (
                user_context.file_sources
                if user_context
                else ConfiguredFileSources.from_dict(None, load_stock_plugins=True)
            )
            stream_url_to_file(
                access_url,
                target_path=str(target_path),
                file_sources=file_sources,
                user_context=user_context,
                file_source_opts=opts,
            )
            downloaded = True
            break
        except exceptions.RequestParameterInvalidException:
            continue

    if not downloaded:
        unimplemented_access_types = [m["type"] for m in access_methods]
        raise _not_implemented(drs_uri, f"that is fetched via unimplemented types ({unimplemented_access_types})")



import abc
import itertools
import re
from typing import (
    List,
    Type,
)

import yaml

from galaxy.util.resources import resource_string


def get_rules_specification():
    return yaml.safe_load(resource_string(__package__, "rules_dsl_spec.yml"))


def _ensure_rule_contains_keys(rule, keys):
    for key, instance_class in keys.items():
        if key not in rule:
            raise ValueError(f"Rule of type [{rule['type']}] does not contain key [{key}].")
        value = rule[key]
        if not isinstance(value, instance_class):
            raise ValueError(f"Rule of type [{rule['type']}] does not contain correct value type for key [{key}].")


def _ensure_key_value_in(rule, key, values):
    value = rule[key]
    if value not in values:
        raise ValueError(f"Invalid value [{value}] for [{key}] encountered.")


def _ensure_valid_pattern(expression):
    re.compile(expression)


def apply_regex(regex, target, data, replacement=None, group_count=None):
    pattern = re.compile(regex)

    def new_row(row):
        source = row[target]
        if replacement is None:
            match = pattern.search(source)
            if not match:
                raise Exception(f"Problem applying regular expression [{regex}] to [{source}].")

            if group_count:
                if len(match.groups()) != group_count:
                    raise Exception("Problem applying regular expression, wrong number of groups found.")

                result = row + list(match.groups())
            else:
                result = row + [match.group(0)]
        else:
            result = row + [pattern.search(source).expand(replacement)]

        return result

    new_data = list(map(new_row, data))
    return new_data


class BaseRuleDefinition(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def rule_type(self):
        """Short string describing type of rule (plugin class) to use."""

    @abc.abstractmethod
    def validate_rule(self, rule):
        """Validate dictified rule definition of this type."""

    @abc.abstractmethod
    def apply(self, rule, data, sources):
        """Apply validated, dictified rule definition to supplied data."""


class AddColumnMetadataRuleDefinition(BaseRuleDefinition):
    rule_type = "add_column_metadata"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(rule, {"value": str})

    def apply(self, rule, data, sources):
        rule_value = rule["value"]
        if rule_value.startswith("identifier"):
            identifier_index = int(rule_value[len("identifier") :])

            new_rows = []
            for index, row in enumerate(data):
                new_rows.append(row + [sources[index]["identifiers"][identifier_index]])

        elif rule_value == "tags":

            def sorted_tags(index):
                tags = sorted(sources[index]["tags"])
                return [",".join(tags)]

            new_rows = []
            for index, row in enumerate(data):
                new_rows.append(row + sorted_tags(index))

        return new_rows, sources


class AddColumnGroupTagValueRuleDefinition(BaseRuleDefinition):
    rule_type = "add_column_group_tag_value"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(rule, {"value": str})

    def apply(self, rule, data, sources):
        rule_value = rule["value"]
        tag_prefix = f"group:{rule_value}:"

        new_rows = []
        for index, row in enumerate(data):
            group_tag_value = None
            source = sources[index]
            tags = source["tags"]
            for tag in sorted(tags):
                if tag.startswith(tag_prefix):
                    group_tag_value = tag[len(tag_prefix) :]
                    break

            if group_tag_value is None:
                group_tag_value = rule.get("default_value", "")

            new_rows.append(row + [group_tag_value])

        return new_rows, sources


class AddColumnConcatenateRuleDefinition(BaseRuleDefinition):
    rule_type = "add_column_concatenate"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(rule, {"target_column_0": int, "target_column_1": int})

    def apply(self, rule, data, sources):
        column_0 = rule["target_column_0"]
        column_1 = rule["target_column_1"]

        new_rows = []
        for row in data:
            new_rows.append(row + [row[column_0] + row[column_1]])

        return new_rows, sources


class AddColumnBasenameRuleDefinition(BaseRuleDefinition):
    rule_type = "add_column_basename"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(rule, {"target_column": int})

    def apply(self, rule, data, sources):
        column = rule["target_column"]
        re = r"[^/]*$"
        return apply_regex(re, column, data), sources


class AddColumnRegexRuleDefinition(BaseRuleDefinition):
    rule_type = "add_column_regex"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(rule, {"target_column": int, "expression": str})
        _ensure_valid_pattern(rule["expression"])

    def apply(self, rule, data, sources):
        target = rule["target_column"]
        expression = rule["expression"]
        replacement = rule.get("replacement")
        group_count = rule.get("group_count")

        return apply_regex(expression, target, data, replacement, group_count), sources


class AddColumnRownumRuleDefinition(BaseRuleDefinition):
    rule_type = "add_column_rownum"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(rule, {"start": int})

    def apply(self, rule, data, sources):
        start = rule["start"]

        new_rows = []
        for index, row in enumerate(data):
            new_rows.append(row + ["%d" % (index + start)])

        return new_rows, sources


class AddColumnValueRuleDefinition(BaseRuleDefinition):
    rule_type = "add_column_value"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(rule, {"value": str})

    def apply(self, rule, data, sources):
        value = rule["value"]

        new_rows = []
        for row in data:
            new_rows.append(row + [str(value)])

        return new_rows, sources


class AddColumnSubstrRuleDefinition(BaseRuleDefinition):
    rule_type = "add_column_substr"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(
            rule,
            {
                "target_column": int,
                "length": int,
                "substr_type": str,
            },
        )
        _ensure_key_value_in(rule, "substr_type", ["keep_prefix", "drop_prefix", "keep_suffix", "drop_suffix"])

    def apply(self, rule, data, sources):
        target = rule["target_column"]
        length = rule["length"]
        substr_type = rule["substr_type"]

        def new_row(row):
            original_value = row[target]
            start = 0
            end = len(original_value)

            if substr_type == "keep_prefix":
                end = length
            elif substr_type == "drop_prefix":
                start = length
            elif substr_type == "keep_suffix":
                start = end - length
                if start < 0:
                    start = 0
            else:
                end = end - length
                if end < 0:
                    end = 0

            return row + [original_value[start:end]]

        return list(map(new_row, data)), sources


class RemoveColumnsRuleDefinition(BaseRuleDefinition):
    rule_type = "remove_columns"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(
            rule,
            {
                "target_columns": list,
            },
        )

    def apply(self, rule, data, sources):
        target_columns = rule["target_columns"]

        def new_row(row):
            new = []
            for index, val in enumerate(row):
                if index not in target_columns:
                    new.append(val)
            return new

        return list(map(new_row, data)), sources


def _filter_index(func, iterable):
    result = []
    for index, x in enumerate(iterable):
        if func(index):
            result.append(x)

    return result


class AddFilterRegexRuleDefinition(BaseRuleDefinition):
    rule_type = "add_filter_regex"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(
            rule,
            {
                "target_column": int,
                "invert": bool,
                "expression": str,
            },
        )
        _ensure_valid_pattern(rule["expression"])

    def apply(self, rule, data, sources):
        target_column = rule["target_column"]
        invert = rule["invert"]
        regex = rule["expression"]

        def _filter(index):
            row = data[index]
            val = row[target_column]
            pattern = re.compile(regex)
            return not invert if pattern.search(val) else invert

        return _filter_index(_filter, data), _filter_index(_filter, sources)


class AddFilterCountRuleDefinition(BaseRuleDefinition):
    rule_type = "add_filter_count"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(
            rule,
            {
                "count": int,
                "invert": bool,
                "which": str,
            },
        )
        _ensure_key_value_in(rule, "which", ["first", "last"])

    def apply(self, rule, data, sources):
        num_rows = len(data)
        invert = rule["invert"]
        n = rule["count"]
        which = rule["which"]

        def _filter(index):
            if which == "first":
                matches = index >= n
            else:
                matches = index < (num_rows - n)
            return not invert if matches else invert

        return _filter_index(_filter, data), _filter_index(_filter, sources)


class AddFilterEmptyRuleDefinition(BaseRuleDefinition):
    rule_type = "add_filter_empty"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(rule, {"target_column": int, "invert": bool})

    def apply(self, rule, data, sources):
        invert = rule["invert"]
        target_column = rule["target_column"]

        def _filter(index):
            non_empty = len(data[index][target_column]) != 0
            return not invert if non_empty else invert

        return _filter_index(_filter, data), _filter_index(_filter, sources)


class AddFilterMatchesRuleDefinition(BaseRuleDefinition):
    rule_type = "add_filter_matches"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(
            rule,
            {
                "target_column": int,
                "invert": bool,
                "value": str,
            },
        )

    def apply(self, rule, data, sources):
        invert = rule["invert"]
        target_column = rule["target_column"]
        value = rule["value"]

        def _filter(index):
            row = data[index]
            val = row[target_column]
            return not invert if val == value else invert

        return _filter_index(_filter, data), _filter_index(_filter, sources)


class AddFilterCompareRuleDefinition(BaseRuleDefinition):
    rule_type = "add_filter_compare"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(
            rule,
            {
                "target_column": int,
                "value": int,
                "compare_type": str,
            },
        )
        _ensure_key_value_in(
            rule, "compare_type", ["less_than", "less_than_equal", "greater_than", "greater_than_equal"]
        )

    def apply(self, rule, data, sources):
        target_column = rule["target_column"]
        value = rule["value"]
        compare_type = rule["compare_type"]

        def _filter(index):
            row = data[index]
            target_value = float(row[target_column])
            if compare_type == "less_than":
                matches = target_value < value
            elif compare_type == "less_than_equal":
                matches = target_value <= value
            elif compare_type == "greater_than":
                matches = target_value > value
            elif compare_type == "greater_than_equal":
                matches = target_value >= value

            return matches

        return _filter_index(_filter, data), _filter_index(_filter, sources)


class SortRuleDefinition(BaseRuleDefinition):
    rule_type = "sort"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(
            rule,
            {
                "target_column": int,
                "numeric": bool,
            },
        )

    def apply(self, rule, data, sources):
        target = rule["target_column"]
        numeric = rule["numeric"]

        sortable = zip(data, sources)

        def sort_func(item):
            a_val = item[0][target]
            if numeric:
                a_val = float(a_val)
            return a_val

        sorted_data = sorted(sortable, key=sort_func)

        new_data = []
        new_sources = []

        for row, source in sorted_data:
            new_data.append(row)
            new_sources.append(source)

        return new_data, new_sources


class SwapColumnsRuleDefinition(BaseRuleDefinition):
    rule_type = "swap_columns"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(
            rule,
            {
                "target_column_0": int,
                "target_column_1": int,
            },
        )

    def apply(self, rule, data, sources):
        target_column_0 = rule["target_column_0"]
        target_column_1 = rule["target_column_1"]

        def new_row(row):
            row_copy = row[:]
            row_copy[target_column_0] = row[target_column_1]
            row_copy[target_column_1] = row[target_column_0]
            return row_copy

        return list(map(new_row, data)), sources


class SplitColumnsRuleDefinition(BaseRuleDefinition):
    rule_type = "split_columns"

    def validate_rule(self, rule):
        _ensure_rule_contains_keys(
            rule,
            {
                "target_columns_0": list,
                "target_columns_1": list,
            },
        )

    def apply(self, rule, data, sources):
        target_columns_0 = rule["target_columns_0"]
        target_columns_1 = rule["target_columns_1"]

        def split_row(row):
            new_row_0 = []
            new_row_1 = []
            for index, el in enumerate(row):
                if index in target_columns_0:
                    new_row_0.append(el)
                elif index in target_columns_1:
                    new_row_1.append(el)
                else:
                    new_row_0.append(el)
                    new_row_1.append(el)

            return [new_row_0, new_row_1]

        data = flat_map(split_row, data)
        sources = flat_map(lambda x: [x, x], sources)

        return data, sources


def flat_map(f, items):
    return list(itertools.chain.from_iterable(map(f, items)))


class RuleSet:
    def __init__(self, rule_set_as_dict):
        self.raw_rules = rule_set_as_dict["rules"]
        self.raw_mapping = rule_set_as_dict.get("mapping", [])

    @property
    def rules(self):
        return self.raw_rules

    def _rules_with_definitions(self):
        for rule in self.raw_rules:
            yield (rule, RULES_DEFINITIONS[rule["type"]])

    def apply(self, data, sources):
        for rule, rule_definition in self._rules_with_definitions():
            rule_definition.validate_rule(rule)
            data, sources = rule_definition.apply(rule, data, sources)

        return data, sources

    @property
    def has_errors(self):
        errored = False
        try:
            for rule, rule_definition in self._rules_with_definitions():
                rule_definition.validate_rule(rule)
        except Exception:
            errored = True
        return errored

    @property
    def mapping_as_dict(self):
        as_dict = {}
        for mapping in self.raw_mapping:
            as_dict[mapping["type"]] = mapping

        return as_dict

    # Rest of this is generic, things here are Galaxy collection specific, think about about
    # subclass of RuleSet for collection creation.
    @property
    def identifier_columns(self):
        mapping_as_dict = self.mapping_as_dict
        identifier_columns = []
        if "list_identifiers" in mapping_as_dict:
            identifier_columns.extend(mapping_as_dict["list_identifiers"]["columns"])
        if "paired_identifier" in mapping_as_dict:
            identifier_columns.append(mapping_as_dict["paired_identifier"]["columns"][0])

        return identifier_columns

    @property
    def collection_type(self):
        mapping_as_dict = self.mapping_as_dict
        list_columns = mapping_as_dict.get("list_identifiers", {"columns": []})["columns"]
        collection_type = ":".join(map(lambda c: "list", list_columns))
        if "paired_identifier" in mapping_as_dict:
            if collection_type:
                collection_type += ":paired"
            else:
                collection_type = "paired"
        return collection_type

    @property
    def display(self):
        message = "Rules:\n"
        message += "".join(f"- {r}\n" for r in self.raw_rules)
        message += "Column Definitions:\n"
        message += "".join(f"- {m}\n" for m in self.raw_mapping)
        return message


RULES_DEFINITION_CLASSES: List[Type[BaseRuleDefinition]] = [
    AddColumnMetadataRuleDefinition,
    AddColumnGroupTagValueRuleDefinition,
    AddColumnConcatenateRuleDefinition,
    AddColumnBasenameRuleDefinition,
    AddColumnRegexRuleDefinition,
    AddColumnRownumRuleDefinition,
    AddColumnValueRuleDefinition,
    AddColumnSubstrRuleDefinition,
    RemoveColumnsRuleDefinition,
    AddFilterRegexRuleDefinition,
    AddFilterCountRuleDefinition,
    AddFilterEmptyRuleDefinition,
    AddFilterMatchesRuleDefinition,
    AddFilterCompareRuleDefinition,
    SortRuleDefinition,
    SwapColumnsRuleDefinition,
    SplitColumnsRuleDefinition,
]
RULES_DEFINITIONS = {}
for rule_class in RULES_DEFINITION_CLASSES:
    RULES_DEFINITIONS[rule_class.rule_type] = rule_class()



import pickle
from io import StringIO


class AliasUnpickler(pickle.Unpickler):
    def __init__(self, aliases, *args, **kw):
        pickle.Unpickler.__init__(self, *args, **kw)
        self.aliases = aliases

    def find_class(self, module, name):
        module, name = self.aliases.get((module, name), (module, name))
        return pickle.Unpickler.find_class(self, module, name)


class AliasPickleModule:
    def __init__(self, aliases):
        self.aliases = aliases

    def dump(self, obj, fileobj, protocol=0):
        return pickle.dump(obj, fileobj, protocol)

    def dumps(self, obj, protocol=0):
        return pickle.dumps(obj, protocol)

    def load(self, fileobj):
        return AliasUnpickler(self.aliases, fileobj).load()

    def loads(self, string):
        fileobj = StringIO(string)
        return AliasUnpickler(self.aliases, fileobj).load()



import pytest


class TestCase:
    """Partial re-implementation of standard library unittest.TestCase using
    pytest methods
    See https://docs.pytest.org/en/latest/how-to/xunit_setup.html for a
    description of the pytest setup/teardown methods.

    Most assert*() methods of unittest.TestCase are not reimplemented here on
    purpose, normal assert statements should be used instead."""

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def setup_class(cls):
        cls.setUpClass()

    @classmethod
    def tearDownClass(cls):
        pass

    @classmethod
    def teardown_class(cls):
        cls.tearDownClass()

    def setUp(self):
        pass

    def setup_method(self):
        self.setUp()

    def tearDown(self):
        pass

    def teardown_method(self):
        self.tearDown()

    def assertRaises(self, exception):
        return pytest.raises(exception)

    def assertRaisesRegex(self, exception, regex):
        return pytest.raises(exception, match=regex)



"""Provides utilities for working with image files."""

import imghdr
import logging

try:
    import Image as PIL
except ImportError:
    try:
        from PIL import Image as PIL
    except ImportError:
        PIL = None

log = logging.getLogger(__name__)


def image_type(filename):
    fmt = None
    if PIL is not None:
        try:
            im = PIL.open(filename)
            fmt = im.format
            im.close()
        except Exception:
            # We continue to try with imghdr, so this is a rare case of an
            # exception we expect to happen frequently, so we're not logging
            pass
    if not fmt:
        fmt = imghdr.what(filename)
    if fmt:
        return fmt.upper()
    else:
        return False


def check_image_type(filename, types):
    fmt = image_type(filename)
    if fmt in types:
        return True
    return False



import logging
from threading import (
    Event,
    Thread,
)

from galaxy.util import ExecutionTimer

log = logging.getLogger(__name__)


class IntervalTask:
    def __init__(self, func, name="Periodic task", interval=3600, immediate_start=False, time_execution=False):
        """
        Run an arbitrary function `func` every `interval` seconds.

        Set `immediate_start` to True to run `func` when task is started.
        """
        self.func = func
        self.name = name
        self.interval = interval
        self.time_execution = time_execution
        self.immediate_start = immediate_start
        self.event = Event()
        self.thread = Thread(target=self.run, name=self.name, daemon=True)
        self.running = False

    def start(self):
        self.running = True
        self.thread.start()

    def _exec(self):
        if self.time_execution:
            timer = ExecutionTimer()
        self.func()
        if self.time_execution:
            log.debug(f"Executed periodic task {self.name} {timer}")

    def run(self):
        if self.immediate_start:
            self._exec()
        while not self.event.is_set():
            self.event.wait(self.interval)
            if self.running:
                self._exec()

    def shutdown(self):
        self.running = False
        self.event.set()
        self.thread.join(5)



from functools import wraps
from typing import (
    Any,
    Callable,
)
from unittest import SkipTest

import pytest
import requests

from galaxy.util.commands import which


def is_site_up(url: str) -> bool:
    try:
        response = requests.get(url, timeout=10)
        return response.status_code == 200
    except Exception:
        return False


def skip_if_site_down(url: str) -> Callable:
    def method_wrapper(method: Callable):
        @wraps(method)
        def wrapped_method(*args, **kwargs) -> Any:
            if not is_site_up(url):
                raise SkipTest(f"Test depends on [{url}] being up and it appears to be down.")
            return method(*args, **kwargs)

        return wrapped_method

    return method_wrapper


skip_if_github_down = skip_if_site_down("https://github.com/")


def _identity(func: Callable) -> Callable:
    return func


def skip_unless_executable(executable):
    if which(executable):
        return _identity
    return pytest.mark.skip(f"PATH doesn't contain executable {executable}")



"""Simple boolean expression parser and evaluator using pyparsing.

Based on the example: https://github.com/pyparsing/pyparsing/blob/master/examples/simpleBool.py
"""

import logging
from typing import (
    Callable,
    Iterable,
    Optional,
    Set,
)

from pyparsing import (
    alphanums,
    CaselessKeyword,
    infixNotation,
    Keyword,
    opAssoc,
    ParseException,
    ParserElement,
    QuotedString,
    Word,
)

log = logging.getLogger(__name__)

ParserElement.enablePackrat()

# Defines the allowed characters that form a valid token.
# Tokens that don't match this format will raise an exception when found.
DEFAULT_TOKEN_FORMAT = f"{alphanums}_-@."

TRUE = Keyword("True")
FALSE = Keyword("False")
NOT_OP = CaselessKeyword("not")
AND_OP = CaselessKeyword("and")
OR_OP = CaselessKeyword("or")
QUOTED_STRING = QuotedString("'")


class TokenEvaluator:
    """Interface to evaluate a token and determine its boolean value."""

    def evaluate(self, token: str) -> bool:
        """Returns the boolean representation of the given token according to some custom logic."""
        raise NotImplementedError


class BoolOperand:
    """Represents a boolean operand that has a label and a value.

    The value is determined by a custom TokenEvaluator."""

    evaluator: TokenEvaluator

    def __init__(self, token):
        self.label = token[0]
        self.value = self.evaluator.evaluate(token[0])

    def __bool__(self):
        return self.value

    def __str__(self):
        return self.label

    __repr__ = __str__


class BoolBinaryOperation:
    """Base representation of a boolean binary operation."""

    reprsymbol: str
    evalop: Callable[[Iterable[object]], bool]

    def __init__(self, token):
        self.args = token[0][0::2]

    def __str__(self):
        sep = f" {self.reprsymbol} "
        return f"({sep.join(map(str, self.args))})"

    def __bool__(self):
        return self.evalop(bool(a) for a in self.args)  # type: ignore[misc,call-arg]

    __nonzero__ = __bool__


class BoolAnd(BoolBinaryOperation):
    """Represents the `AND` boolean operation."""

    reprsymbol = "&"
    evalop = all


class BoolOr(BoolBinaryOperation):
    """Represents the `OR` boolean operation."""

    reprsymbol = "|"
    evalop = any


class BoolNot:
    """Represents the `NOT` boolean operation."""

    def __init__(self, token):
        self.arg = token[0][1]

    def __bool__(self):
        v = bool(self.arg)
        return not v

    def __str__(self):
        return f"~{self.arg}"

    __repr__ = __str__


class BooleanExpressionEvaluator:
    """Boolean logic parser that can evaluate an expression using a particular TokenEvaluator.

    Supports AND, OR and NOT operator including parentheses to override operator precedences.

    You can pass in different TokenEvaluator implementations to customize how the tokens (or variables) are
    converted to a boolean value when evaluating the expression."""

    def __init__(self, evaluator: TokenEvaluator, token_format: Optional[str] = None) -> None:
        """Initializes the expression evaluator.

        :param evaluator: The custom TokenEvaluator used to transform any token into a boolean.
        :type evaluator:  TokenEvaluator

        :param token_format: A string of all allowed characters used to form a valid token, defaults to None.
                             The default value (None) will use DEFAULT_TOKEN_FORMAT which means the allowed characters are ``[A-Za-z0-9_-@.]``.
        :type token_format:  Optional[str]
        """
        action = BoolOperand
        action.evaluator = evaluator
        boolOperand = TRUE | FALSE | QUOTED_STRING | Word(token_format or DEFAULT_TOKEN_FORMAT)
        boolOperand.setParseAction(action)
        self.boolExpr: ParserElement = infixNotation(
            boolOperand,
            [
                (NOT_OP, 1, opAssoc.RIGHT, BoolNot),
                (AND_OP, 2, opAssoc.LEFT, BoolAnd),
                (OR_OP, 2, opAssoc.LEFT, BoolOr),
            ],
        )

    def evaluate_expression(self, expr: str) -> bool:
        """Given an expression it gets evaluated to True or False using boolean logic."""
        try:
            res = self.boolExpr.parseString(expr, parseAll=True)[0]
            return bool(res)
        except ParseException as e:
            log.error(f"BooleanExpressionEvaluator unable to evaluate expression => {expr}", exc_info=e)
            raise e

    @classmethod
    def is_valid_expression(cls, expr: str) -> bool:
        """Tries to evaluate the given boolean expression and returns True if it is valid or
        False if it has syntax or gramatical errors."""
        try:
            evaluator = BooleanExpressionEvaluator(ValidationOnlyTokenEvaluator())
            evaluator.evaluate_expression(expr)
            return True
        except ParseException:
            return False


class TokenContainedEvaluator(TokenEvaluator):
    """Implements the TokenEvaluator interface to determine if a token is contained
    in a particular list of tokens."""

    def __init__(self, tokens: Set[str]) -> None:
        """Initializes the token evaluator with the set of tokens that will evaluate to `True`.

        :param tokens: The list of tokens that should be evaluated to True.
        :type tokens: List[str]
        """
        self.tokens = tokens or set()

    def evaluate(self, token: str) -> bool:
        return token in self.tokens


class ValidationOnlyTokenEvaluator(TokenEvaluator):
    """Simple TokenEvaluator that always evaluates to True for valid tokens.

    This is only useful for validation purposes, do NOT use it for real expression evaluations."""

    def evaluate(self, token: str) -> bool:
        return True



import copy
import json
import logging
import math
import random
import string
from collections.abc import (
    Iterable,
    Mapping,
    Sequence,
)
from decimal import Decimal

from . import unicodify

__all__ = ("safe_dumps", "validate_jsonrpc_request", "validate_jsonrpc_response", "jsonrpc_request", "jsonrpc_response")

log = logging.getLogger(__name__)

to_json_string = json.dumps
from_json_string = json.loads


def swap_inf_nan(val):
    """
    This takes an arbitrary object and preps it for jsonifying safely, templating Inf/NaN and
    casting Decimal instances as strings.
    """
    if isinstance(val, str):
        # basestring first, because it's a sequence and would otherwise get caught below.
        return val
    elif isinstance(val, Sequence):
        return [swap_inf_nan(v) for v in val]
    elif isinstance(val, Mapping):
        return {swap_inf_nan(k): swap_inf_nan(v) for (k, v) in val.items()}
    elif isinstance(val, float):
        if math.isnan(val):
            return "__NaN__"
        elif val == math.inf:
            return "__Infinity__"
        elif val == -math.inf:
            return "__-Infinity__"
        else:
            return val
    elif isinstance(val, Decimal):
        return str(val)
    else:
        return val


def safe_loads(arg):
    """
    This is a wrapper around loads that returns the parsed value instead of
    raising a value error. It also avoids autoconversion of non-iterables
    i.e numeric and boolean values.
    """
    try:
        loaded = json.loads(arg)
        if loaded is not None and not isinstance(loaded, Iterable):
            loaded = arg
    except (TypeError, ValueError):
        loaded = copy.deepcopy(arg)
    return loaded


def safe_dumps(obj, **kwargs):
    """
    This is a wrapper around dumps that encodes Infinity and NaN values.  It's a
    fairly rare case (which will be low in request volume).  Basically, we tell
    json.dumps to blow up if it encounters Infinity/NaN, or Decimal values
    and we 'fix' it before re-encoding.
    """
    try:
        dumped = json.dumps(obj, allow_nan=False, **kwargs)
    except (ValueError, TypeError):
        obj = swap_inf_nan(obj)
        dumped = json.dumps(obj, allow_nan=False, **kwargs)
    if kwargs.get("escape_closing_tags", True):
        return dumped.replace("</", "<\\/")
    return dumped


def safe_dumps_formatted(obj):
    """Attempt to format an object for display.

    If serialization fails, the object's string representation will be returned instead.
    """
    try:
        return safe_dumps(obj, sort_keys=True, indent=4, separators=(",", ": "))
    except TypeError:
        log.warning("JSON serialization failed for object: %s", str(obj))
        return str(obj)


# Methods for handling JSON-RPC


def validate_jsonrpc_request(request, regular_methods, notification_methods):
    try:
        request = json.loads(request)
    except Exception as e:
        return (
            False,
            request,
            jsonrpc_response(id=None, error=dict(code=-32700, message="Parse error", data=unicodify(e))),
        )
    try:
        assert (
            "jsonrpc" in request
        ), 'This server requires JSON-RPC 2.0 and no "jsonrpc" member was sent with the Request object as per the JSON-RPC 2.0 Specification.'
        assert (
            request["jsonrpc"] == "2.0"
        ), f"Requested JSON-RPC version \"{request['jsonrpc']}\" != required version \"2.0\"."
        assert "method" in request, 'No "method" member was sent with the Request object'
    except AssertionError as e:
        return (
            False,
            request,
            jsonrpc_response(request=request, error=dict(code=-32600, message="Invalid Request", data=unicodify(e))),
        )
    try:
        assert request["method"] in (regular_methods + notification_methods)
    except AssertionError:
        return (
            False,
            request,
            jsonrpc_response(
                request=request,
                error=dict(
                    code=-32601,
                    message="Method not found",
                    data=f"Valid methods are: {', '.join(regular_methods + notification_methods)}",
                ),
            ),
        )
    try:
        if request["method"] in regular_methods:
            assert (
                "id" in request
            ), f"No \"id\" member was sent with the Request object and the requested method \"{request['method']}\" is not a notification method"
    except AssertionError as e:
        return (
            False,
            request,
            jsonrpc_response(request=request, error=dict(code=-32600, message="Invalid Request", data=unicodify(e))),
        )
    return True, request, None


def validate_jsonrpc_response(response, id=None):
    try:
        response = json.loads(response)
    except Exception as e:
        log.error("Response was not valid JSON: %s", unicodify(e))
        log.debug("Response was: %s", response)
        return False, response
    try:
        assert (
            "jsonrpc" in response
        ), 'This server requires JSON-RPC 2.0 and no "jsonrpc" member was sent with the Response object as per the JSON-RPC 2.0 Specification.'
        assert (
            "result" in response or "error" in response
        ), 'Neither of "result" or "error" members were sent with the Response object.'
        if "error" in response:
            assert int(
                response["error"]["code"]
            ), 'The "code" member of the "error" object in the Response is missing or not an integer.'
            assert "message" in response, 'The "message" member of the "error" object in the Response is missing.'
    except Exception:
        log.exception("Response was not valid JSON-RPC")
        log.debug(f"Response was: {response}")
        return False, response
    if id is not None:
        try:
            assert "id" in response and response["id"] == id
        except Exception:
            log.error(f"The response id \"{response['id']}\" does not match the request id \"{id}\"")
            return False, response
    return True, response


def jsonrpc_request(method, params=None, id=None, jsonrpc="2.0"):
    if method is None:
        log.error('jsonrpc_request(): "method" parameter cannot be None')
        return None
    request = dict(jsonrpc=jsonrpc, method=method)
    if params:
        request["params"] = params
    if id is not None and id is True:
        request["id"] = "".join(random.choice(string.hexdigits) for i in range(16))
    elif id is not None:
        request["id"] = id
    return request


def jsonrpc_response(request=None, id=None, result=None, error=None, jsonrpc="2.0"):
    if result:
        rval = dict(jsonrpc=jsonrpc, result=result)
    elif error:
        rval = dict(jsonrpc=jsonrpc, error=error)
    else:
        msg = 'jsonrpc_response() called with out a "result" or "error" parameter'
        log.error(msg)
        rval = dict(jsonrpc=jsonrpc, error=msg)
    if id is not None:
        rval["id"] = id
    elif request is not None and "id" in request:
        rval["id"] = request["id"]
    return rval



"""Common markdown formatting helpers for per-datatype rendering."""


def literal_via_fence(content):
    return "\n%s\n" % "\n".join(f"    {line}" for line in content.splitlines())


def indicate_data_truncated():
    return "\n**Warning:** The above data has been truncated to be embedded in this document.\n\n"


def pre_formatted_contents(markdown):
    return f"<pre>{markdown}</pre>"



"""Return various facts for string formatting.
"""
import socket
from collections.abc import MutableMapping


class Facts(MutableMapping):
    """A dict-like object that evaluates values at access time."""

    def __init__(self, config=None, **kwargs):
        config = config or {}
        self.__dict__ = {}
        self.__set_defaults(config)
        self.__set_config(config)
        self.__dict__.update(dict(**kwargs))

    def __set_defaults(self, config):
        # config here may be a Galaxy config object, or it may just be a dict
        defaults = {
            "server_name": lambda: config.get("base_server_name", "main"),
            "server_id": None,
            "instance_id": None,
            "pool_name": None,
            "fqdn": lambda: socket.getfqdn(),
            "hostname": lambda: socket.gethostname().split(".", 1)[0],
        }
        self.__dict__.update(defaults)

    def __set_config(self, config):
        if config is not None:
            for name in dir(config):
                if not name.startswith("_") and isinstance(getattr(config, name), str):
                    self.__dict__[f"config_{name}"] = lambda name=name: getattr(config, name)

    def __getitem__(self, key):
        item = self.__dict__.__getitem__(key)
        if callable(item):
            return item()
        else:
            return item

    # Other methods pass through to the corresponding dict methods

    def __setitem__(self, key, value):
        return self.__dict__.__setitem__(key, value)

    def __delitem__(self, key):
        return self.__dict__.__delitem__(key)

    def __iter__(self):
        return self.__dict__.__iter__()

    def __len__(self):
        return self.__dict__.__len__()

    def __str__(self):
        return self.__dict__.__str__()

    def __repr__(self):
        return self.__dict__.__repr__()


def get_facts(config=None, **kwargs):
    return Facts(config=config, **kwargs)



"""Provide a consistent interface into and utilities for importlib file resources.
"""
try:
    from importlib.abc import Traversable  # type: ignore[attr-defined]
    from importlib.resources import files  # type: ignore[attr-defined]
except ImportError:
    # Python < 3.9
    from importlib_resources import files  # type: ignore[no-redef]
    from importlib_resources.abc import Traversable  # type: ignore[no-redef]


def resource_path(package_or_requirement, resource_name):
    """
    Return specified resource as a string.

    Replacement function for pkg_resources.resource_string, but returns unicode string instead of bytestring.
    """
    return files(package_or_requirement).joinpath(resource_name)


def resource_string(package_or_requirement, resource_name) -> str:
    """
    Return specified resource as a string.

    Replacement function for pkg_resources.resource_string, but returns unicode string instead of bytestring.
    """
    return resource_path(package_or_requirement, resource_name).read_text()


__all__ = (
    "files",
    "resource_string",
    "resource_path",
    "Traversable",
)



import logging
import os
import tempfile
from typing import (
    Optional,
    Tuple,
)

from galaxy.util import (
    etree,
    parse_xml as galaxy_parse_xml,
    unicodify,
    xml_to_string,
)
from galaxy.util.path import StrPath

log = logging.getLogger(__name__)


def create_and_write_tmp_file(elem: etree.Element) -> str:
    tmp_str = xml_to_string(elem, pretty=True)
    with tempfile.NamedTemporaryFile(prefix="tmp-toolshed-cawrf", delete=False) as fh:
        tmp_filename = fh.name
    with open(tmp_filename, mode="w", encoding="utf-8") as fh:
        fh.write(tmp_str)
    return tmp_filename


def parse_xml(file_name: StrPath, check_exists=True) -> Tuple[Optional[etree.ElementTree], str]:
    """Returns a parsed xml tree with comments intact."""
    error_message = ""
    if check_exists and not os.path.exists(file_name):
        return None, f"File does not exist {str(file_name)}"
    try:
        tree = galaxy_parse_xml(file_name, remove_comments=False, strip_whitespace=False)
    except OSError:
        raise
    except Exception as e:
        error_message = f"Exception attempting to parse {str(file_name)}: {unicodify(e)}"
        log.exception(error_message)
        return None, error_message
    return tree, error_message


__all__ = (
    "create_and_write_tmp_file",
    "parse_xml",
)



# TODO: this is largely copied from galaxy.tool_util.toolbox.galaxy and generalized, the tool-oriented watchers in that
# module should probably be updated to use this where possible


import logging
import os.path
import time

try:
    from watchdog.events import FileSystemEventHandler
    from watchdog.observers import Observer
    from watchdog.observers.polling import PollingObserver

    can_watch = True
except ImportError:
    Observer = None
    FileSystemEventHandler = object
    PollingObserver = None
    can_watch = False

from galaxy.util.hash_util import md5_hash_file

log = logging.getLogger(__name__)


def get_observer_class(config_name, config_value, default, monitor_what_str):
    """ """
    config_value = config_value or default
    config_value = str(config_value).lower()
    if config_value in ("true", "yes", "on", "auto"):
        expect_observer = True
        observer_class = Observer
    elif config_value == "polling":
        expect_observer = True
        observer_class = PollingObserver
    elif config_value in ("false", "no", "off"):
        expect_observer = False
        observer_class = None
    else:
        message = f"Unrecognized value for {config_name} config option: {config_value}"
        raise Exception(message)

    if expect_observer and observer_class is None:
        message = f"Watchdog library unavailable, cannot monitor {monitor_what_str}."
        if config_value == "auto":
            log.info(message)
        else:
            raise Exception(message)

    return observer_class


def get_watcher(
    config, config_name, default="False", monitor_what_str=None, watcher_class=None, event_handler_class=None, **kwargs
):
    config_value = getattr(config, config_name, None)
    observer_class = get_observer_class(config_name, config_value, default=default, monitor_what_str=monitor_what_str)
    if observer_class is not None:
        watcher_class = watcher_class or Watcher
        event_handler_class = event_handler_class or EventHandler
        return watcher_class(observer_class, event_handler_class, **kwargs)
    else:
        return NullWatcher()


class BaseWatcher:
    def __init__(self, observer_class, event_handler_class, **kwargs):
        self.observer = None
        self.observer_class = observer_class
        self.event_handler = event_handler_class(self)
        self.monitored_dirs = {}

    def start(self):
        if self.observer is None:
            self.observer = self.observer_class()
            self.observer.start()
            self.resume_watching()

    def monitor(self, dir_path, recursive=False):
        self.monitored_dirs[dir_path] = recursive
        if self.observer is not None:
            self.observer.schedule(self.event_handler, dir_path, recursive=recursive)

    def resume_watching(self):
        for dir_path, recursive in self.monitored_dirs.items():
            self.monitor(dir_path, recursive)

    def shutdown(self):
        if self.observer is not None:
            self.observer.stop()
            self.observer.join()
            self.observer = None


class Watcher(BaseWatcher):
    def __init__(self, observer_class, event_handler_class, **kwargs):
        super().__init__(observer_class, event_handler_class, **kwargs)
        self.path_hash = {}
        self.file_callbacks = {}
        self.dir_callbacks = {}
        self.ignore_extensions = {}
        self.require_extensions = {}
        self.event_handler = event_handler_class(self)

    def watch_file(self, file_path, callback=None):
        file_path = os.path.abspath(file_path)
        dir_path = os.path.dirname(file_path)
        if dir_path not in self.monitored_dirs:
            if callback is not None:
                self.file_callbacks[file_path] = callback
            self.monitor(dir_path)
            log.debug("Watching for changes to file: %s", file_path)

    def watch_directory(
        self, dir_path, callback=None, recursive=False, ignore_extensions=None, require_extensions=None
    ):
        dir_path = os.path.abspath(dir_path)
        if dir_path not in self.monitored_dirs:
            if callback is not None:
                self.dir_callbacks[dir_path] = callback
            if ignore_extensions:
                self.ignore_extensions[dir_path] = ignore_extensions
            if require_extensions:
                self.require_extensions[dir_path] = require_extensions
            self.monitor(dir_path, recursive=recursive)
            log.debug("Watching for changes in directory%s: %s", " (recursively)" if recursive else "", dir_path)


class EventHandler(FileSystemEventHandler):
    def __init__(self, watcher):
        self.watcher = watcher

    # this effectively excludes on_opened and on_closed
    def on_moved(self, event):
        self._handle(event)

    def on_created(self, event):
        self._handle(event)

    def on_deleted(self, event):
        self._handle(event)

    def on_modified(self, event):
        self._handle(event)

    def _extension_check(self, key, path):
        required_extensions = self.watcher.require_extensions.get(key)
        if required_extensions:
            return any(filter(path.endswith, required_extensions))
        return not any(filter(path.endswith, self.watcher.ignore_extensions.get(key, [])))

    def _handle(self, event):
        # modified events will only have src path, move events will
        # have dest_path and src_path but we only care about dest. So
        # look at dest if it exists else use src.
        path = getattr(event, "dest_path", None) or event.src_path
        path = os.path.abspath(path)
        callback = self.watcher.file_callbacks.get(path)
        if os.path.basename(path).startswith("."):
            return
        if callback:
            ext_ok = self._extension_check(path, path)
        else:
            # reversed sort for getting the most specific dir first
            for key in reversed(sorted(self.watcher.dir_callbacks.keys())):
                if os.path.commonprefix([path, key]) == key:
                    callback = self.watcher.dir_callbacks[key]
                    ext_ok = self._extension_check(key, path)
                    break
        if not callback or not ext_ok:
            return
        cur_hash = md5_hash_file(path)
        if cur_hash:
            if self.watcher.path_hash.get(path) == cur_hash:
                return
            else:
                time.sleep(0.5)
                if cur_hash != md5_hash_file(path):
                    # We're still modifying the file, it'll be picked up later
                    return
                self.watcher.path_hash[path] = cur_hash
                callback(path=path)


class NullWatcher:
    def start(self):
        pass

    def shutdown(self):
        pass

    def watch_file(self, *args, **kwargs):
        pass

    def watch_directory(self, *args, **kwargs):
        pass



import functools
import operator

from galaxy import util


# Utility methods for specifing maps.
def to_str_or_none(value):
    if value is None:
        return None
    else:
        return str(value)


def to_bool_or_none(value):
    return util.string_as_bool_or_none(value)


def to_bool(value):
    return util.asbool(value)


def to_float_or_none(value):
    if value is None:
        return None
    else:
        return float(value)


# Utility methods for specifing valid...
def is_in(*args):
    return functools.partial(operator.contains, args)



import logging
import threading

from .sleeper import Sleeper

log = logging.getLogger(__name__)

DEFAULT_MONITOR_THREAD_JOIN_TIMEOUT = 5


class Monitors:
    def _init_monitor_thread(self, name, target_name=None, target=None, start=False, config=None):
        self.monitor_join_sleep = getattr(config, "monitor_thread_join_timeout", DEFAULT_MONITOR_THREAD_JOIN_TIMEOUT)
        self.monitor_join = self.monitor_join_sleep > 0
        self.monitor_running = True

        if target is not None:
            assert target_name is None
            monitor_func = target
        else:
            target_name = target_name or "monitor"
            monitor_func = getattr(self, target_name)
        self.sleeper = Sleeper()
        self.monitor_thread = threading.Thread(name=name, target=monitor_func)
        self.monitor_thread.daemon = True
        self._start = start
        self.start_monitoring()

    def _init_noop_monitor(self):
        self.sleeper = None
        self.monitor_join = False

    def start_monitoring(self):
        if self._start:
            self.monitor_thread.start()

    def stop_monitoring(self):
        self.monitor_running = False

    def _monitor_sleep(self, sleep_amount):
        self.sleeper.sleep(sleep_amount)

    def shutdown_monitor(self):
        self.stop_monitoring()
        if self.sleeper is not None:
            self.sleeper.wake()
        if self.monitor_join:
            log.debug("Joining monitor thread")
            self.monitor_thread.join(self.monitor_join_sleep)



from typing import (
    Any,
    TYPE_CHECKING,
)

if TYPE_CHECKING:

    class HasDynamicProperties:
        def __getattr__(self, property: str) -> Any:
            return object()

else:
    HasDynamicProperties = object



from collections import UserDict
from collections.abc import (
    ItemsView,
    MutableMapping,
)
from typing import (
    Any,
    Optional,
)

from boltons.iterutils import remap


def enter(path, key, value):
    if isinstance(value, MutableMapping):
        return value.__class__(), ItemsView(value)
    else:
        return value, False


class TreeDict(UserDict):
    """
    Dictionary that inserts its own keys in a parent dictionary.
    """

    def __init__(self, dict=None, **kwargs):
        self._parent_data: Optional[TreeDict] = None
        self._injected_data = {}
        super().__init__(dict, **kwargs)

    def clean_copy(self):
        """
        Copy without injected data.
        """

        def strip_tree_dict(path, key, value):
            if isinstance(value, TreeDict):
                value = value.data
            return key, value

        return remap(self.data, strip_tree_dict, enter=enter)

    def __getitem__(self, key: Any) -> Any:
        if key in self.data:
            return super().__getitem__(key)
        else:
            return self._injected_data[key]

    def __contains__(self, key: object) -> bool:
        if super().__contains__(key):
            return True
        return key in self._injected_data

    def __setitem__(self, key: Any, item: Any) -> None:
        if isinstance(item, MutableMapping):
            # We're not doing item = TreeDict(item) because we want to record the keys in _parent_data
            _item = TreeDict()
            _item._parent_data = self
            _item.update(item)
            item = _item
        current_parent_data = self._parent_data
        while current_parent_data is not None and key != "__current_case__":
            if (
                key not in current_parent_data
                or key == "input"
                and key in current_parent_data
                and callable(current_parent_data[key])
            ):
                current_parent_data._injected_data[key] = item
            current_parent_data = current_parent_data._parent_data
        return super().__setitem__(key, item)



import os
from copy import deepcopy
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)

from galaxy.util import (
    Element,
    ElementTree,
    parse_xml,
)
from galaxy.util.path import StrPath

REQUIRED_PARAMETER = object()


def load_with_references(path: StrPath) -> Tuple[ElementTree, Optional[List[str]]]:
    """Load XML documentation from file system and preprocesses XML macros.

    Return the XML representation of the expanded tree and paths to
    referenced files that were imported (macros).
    """
    tree = raw_xml_tree(path)
    root = tree.getroot()

    macros_el = _macros_el(root)
    if macros_el is None:
        return tree, []

    macros: Dict[str, List[Element]] = {}
    macro_paths = _import_macros(macros_el, path, macros)
    macros_el.clear()

    # Collect tokens
    tokens = {}
    for m in macros.get("token", []):
        tokens[m.get("name")] = m.text or ""
    tokens = expand_nested_tokens(tokens)

    # Expand xml macros
    macro_dict = {}
    for m in macros.get("xml", []):
        macro_dict[m.get("name")] = XmlMacroDef(m)
    _expand_macros([root], macro_dict, tokens)

    # reinsert template macro which are used during tool execution
    for m in macros.get("template", []):
        macros_el.append(m)
    _expand_tokens_for_el(root, tokens)
    return tree, macro_paths


def load(path: StrPath) -> ElementTree:
    tree, _ = load_with_references(path)
    return tree


def template_macro_params(root):
    """
    Look for template macros and populate param_dict (for cheetah)
    with these.
    """
    param_dict = {}
    macros_el = _macros_el(root)
    macro_dict = _macros_of_type(macros_el, "template", lambda el: el.text)
    for key, value in macro_dict.items():
        param_dict[key] = value
    return param_dict


def raw_xml_tree(path: StrPath) -> ElementTree:
    """Load raw (no macro expansion) tree representation of XML represented
    at the specified path.
    """
    tree = parse_xml(path, strip_whitespace=False, remove_comments=True)
    return tree


def imported_macro_paths(root):
    macros_el = _macros_el(root)
    return _imported_macro_paths_from_el(macros_el)


def _import_macros(macros_el, path, macros) -> Optional[List[str]]:
    """
    root the parsed XML tree
    path the path to the main xml document
    """
    xml_base_dir = os.path.dirname(path)
    if macros_el is not None:
        macro_paths = _load_macros(macros_el, xml_base_dir, macros)
        # _xml_set_children(macros_el, macro_els)
        return macro_paths
    return None


def _macros_el(root):
    return root.find("macros")


def _macros_of_type(macros_el, type, el_func):
    if macros_el is None:
        return {}
    macro_els = macros_el.findall("macro")
    filtered_els = [(macro_el.get("name"), el_func(macro_el)) for macro_el in macro_els if macro_el.get("type") == type]
    return dict(filtered_els)


def expand_nested_tokens(tokens):
    for token_name in tokens.keys():
        for current_token_name, current_token_value in tokens.items():
            if token_name in current_token_value:
                if token_name == current_token_name:
                    raise Exception(f"Token '{token_name}' cannot contain itself")
                tokens[current_token_name] = current_token_value.replace(token_name, tokens[token_name])
    return tokens


def _expand_tokens(elements, tokens):
    if not tokens or elements is None:
        return

    for element in elements:
        _expand_tokens_for_el(element, tokens)


def _expand_tokens_for_el(element, tokens):
    """
    expand tokens in element and (recursively) in its children
    replacements of text attributes and attribute values are
    possible
    """
    value = element.text
    if value:
        new_value = _expand_tokens_str(element.text, tokens)
        if new_value is not value:
            element.text = new_value
    for key, value in element.attrib.items():
        new_value = _expand_tokens_str(value, tokens)
        if new_value is not value:
            element.attrib[key] = new_value
        new_key = _expand_tokens_str(key, tokens)
        if new_key is not key:
            element.attrib[new_key] = element.attrib[key]
            del element.attrib[key]
    # recursively expand in childrens
    _expand_tokens(list(element), tokens)


def _expand_tokens_str(s, tokens):
    for key, value in tokens.items():
        if key in s:
            s = s.replace(key, value)
    return s


def _expand_macros(elements, macros, tokens, visited=None):
    if not macros and not tokens:
        return

    if visited is None:
        v = []
    else:
        v = visited

    for element in elements:
        while True:
            expand_el = element.find(".//expand")
            if expand_el is None:
                break
            _expand_macro(expand_el, macros, tokens, v)


def _expand_macro(expand_el, macros, tokens, visited):
    macro_name = expand_el.get("macro")
    assert macro_name is not None, "Attempted to expand macro with no 'macro' attribute defined."
    # check for cycles in the nested macro expansion
    assert (
        macro_name not in visited
    ), f"Cycle in nested macros: already expanded {visited} can't expand '{macro_name}' again"
    visited.append(macro_name)

    assert macro_name in macros, f"No macro named {macro_name} found, known macros are {', '.join(macros.keys())}."
    macro_def = macros[macro_name]
    expanded_elements = deepcopy(macro_def.element)
    _expand_yield_statements(expanded_elements, expand_el)

    macro_tokens = macro_def.macro_tokens(expand_el)
    if macro_tokens:
        _expand_tokens(expanded_elements, macro_tokens)

    # Recursively expand contained macros.
    _expand_macros(expanded_elements, macros, tokens, visited)
    _xml_replace(expand_el, expanded_elements)
    del visited[-1]


def _expand_yield_statements(macro_def, expand_el):
    """
    Modifies the macro_def element by replacing
    1. all named yield tags by the content of the corresponding token tags
       - token tags need to be direct children of the expand
       - processed in order of definition of the token tags
    2. all unnamed yield tags by the non-token children of the expand tag
    """
    # replace named yields
    for token_el in expand_el.findall("./token"):
        name = token_el.attrib.get("name", None)
        assert name is not None, "Found unnamed token" + str(token_el.attrib)
        yield_els = [yield_el for yield_el in macro_def.findall(f".//yield[@name='{name}']")]
        assert len(yield_els) > 0, f"No named yield found for named token {name}"
        token_el_children = list(token_el)
        for yield_el in yield_els:
            _xml_replace(yield_el, token_el_children)

    # replace unnamed yields
    yield_els = [yield_el for yield_el in macro_def.findall(".//yield")]
    expand_el_children = [c for c in expand_el if c.tag != "token"]
    for yield_el in yield_els:
        _xml_replace(yield_el, expand_el_children)


def _load_macros(macros_el, xml_base_dir, macros) -> List[str]:
    # Import macros from external files.
    macro_paths = _load_imported_macros(macros_el, xml_base_dir, macros)
    # Load all directly defined macros.
    _load_embedded_macros(macros_el, xml_base_dir, macros)
    return macro_paths


def _load_embedded_macros(macros_el, xml_base_dir, macros):
    if macros_el is None:
        return
    # attribute typed macro
    for macro in macros_el.iterfind("macro"):
        if "type" not in macro.attrib:
            macro.attrib["type"] = "xml"
        try:
            macros[macro.attrib["type"]].append(macro)
        except KeyError:
            macros[macro.attrib["type"]] = [macro]

    # type shortcuts (<xml> is a shortcut for <macro type="xml",
    # likewise for <template>.
    for tag in ["template", "xml", "token"]:
        for macro_el in macros_el.iterfind(tag):
            macro_el.attrib["type"] = tag
            macro_el.tag = "macro"
            try:
                macros[tag].append(macro_el)
            except KeyError:
                macros[tag] = [macro_el]


def _load_imported_macros(macros_el, xml_base_dir, macros) -> List[str]:
    macro_paths = []

    for tool_relative_import_path in _imported_macro_paths_from_el(macros_el):
        import_path = os.path.join(xml_base_dir, tool_relative_import_path)
        macro_paths.append(import_path)
        current_macro_paths = _load_macro_file(import_path, xml_base_dir, macros)
        macro_paths.extend(current_macro_paths)
    return macro_paths


def _imported_macro_paths_from_el(macros_el):
    imported_macro_paths = []
    macro_import_els = []
    if macros_el is not None:
        macro_import_els = macros_el.findall("import")
    for macro_import_el in macro_import_els:
        raw_import_path = macro_import_el.text
        imported_macro_paths.append(raw_import_path)
    return imported_macro_paths


def _load_macro_file(path: StrPath, xml_base_dir, macros) -> List[str]:
    tree = parse_xml(path, strip_whitespace=False)
    root = tree.getroot()
    return _load_macros(root, xml_base_dir, macros)


def _xml_set_children(element, new_children):
    for old_child in element:
        element.remove(old_child)
    for i, new_child in enumerate(new_children):
        element.insert(i, new_child)


def _xml_replace(query, targets):
    parent_el = query.find("..")
    matching_index = -1
    # for index, el in enumerate(parent_el.iter('.')):  ## Something like this for newer implementation
    for index, el in enumerate(list(parent_el)):
        if el == query:
            matching_index = index
            break
    assert matching_index >= 0
    current_index = matching_index
    for target in targets:
        current_index += 1
        parent_el.insert(current_index, deepcopy(target))
    parent_el.remove(query)


class XmlMacroDef:
    """
    representation of a (Galaxy) XML macro

    stores the root element of the macro and the parameters.
    each parameter is represented as pair containing
    - the quote character, default '@'
    - parameter name

    parameter names can be given as comma separated list using the
    `token` attribute or as attributes `token_XXX` (where `XXX` is the name).
    The former option should be used to specify required attributes of the
    macro and the latter for optional attributes if the macro (the value of
    `token_XXX is used as default value).

    TODO: `token_quote` forbids `"quote"` as character name of optional
    parameters
    """

    def __init__(self, el):
        self.element = el
        parameters = {}
        tokens = []
        token_quote = "@"
        for key, value in el.attrib.items():
            if key == "token_quote":
                token_quote = value
            if key == "tokens":
                for token in value.split(","):
                    tokens.append((token, REQUIRED_PARAMETER))
            elif key.startswith("token_"):
                token = key[len("token_") :]
                tokens.append((token, value))
        for name, default in tokens:
            parameters[name] = (token_quote, default)
        self.parameters = parameters

    def macro_tokens(self, expand_el):
        """
        get a dictionary mapping token names to values. The names are the
        parameter names surrounded by the quote character. Values are taken
        from the expand_el if absent default values of optional parameters are
        used.
        """
        tokens = {}
        for key, (wrap_char, default_val) in self.parameters.items():
            token_value = expand_el.attrib.get(key, default_val)
            if token_value is REQUIRED_PARAMETER:
                raise ValueError(f"Failed to expand macro - missing required parameter [{key}].")
            token_name = f"{wrap_char}{key.upper()}{wrap_char}"
            tokens[token_name] = token_value
        return tokens


__all__ = (
    "imported_macro_paths",
    "load",
    "load_with_references",
    "raw_xml_tree",
    "template_macro_params",
)



"""Code obtained from https://github.com/dmfrey/FileLock.

See full license at:

https://github.com/dmfrey/FileLock/blob/master/LICENSE.txt

"""
import errno
import os
import time


class FileLockException(Exception):
    pass


class FileLock:
    """A file locking mechanism that has context-manager support so
    you can use it in a with statement. This should be relatively cross
    compatible as it doesn't rely on msvcrt or fcntl for the locking.
    """

    def __init__(self, file_name, timeout=10, delay=0.05):
        """Prepare the file locker. Specify the file to lock and optionally
        the maximum timeout and the delay between each attempt to lock.
        """
        self.is_locked = False
        full_path = os.path.abspath(file_name)
        self.lockfile = f"{full_path}.lock"
        self.file_name = full_path
        self.timeout = timeout
        self.delay = delay

    def acquire(self):
        """Acquire the lock, if possible. If the lock is in use, it check again
        every `wait` seconds. It does this until it either gets the lock or
        exceeds `timeout` number of seconds, in which case it throws
        an exception.
        """
        start_time = time.time()
        while True:
            try:
                self.fd = os.open(self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
                break
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
                if (time.time() - start_time) >= self.timeout:
                    raise FileLockException("Timeout occurred.")
                time.sleep(self.delay)
        self.is_locked = True

    def release(self):
        """Get rid of the lock by deleting the lockfile.
        When working in a `with` statement, this gets automatically
        called at the end.
        """
        if self.is_locked:
            os.close(self.fd)
            os.unlink(self.lockfile)
            self.is_locked = False

    def __enter__(self):
        """Activated when used in the with statement.
        Should automatically acquire a lock to be used in the with block.
        """
        if not self.is_locked:
            self.acquire()
        return self

    def __exit__(self, type, value, traceback):
        """Activated at the end of the with statement.
        It automatically releases the lock if it isn't locked.
        """
        if self.is_locked:
            self.release()

    def __del__(self):
        """Make sure that the FileLock instance doesn't leave a lockfile
        lying around.
        """
        self.release()



import binascii
import json

from galaxy.util import (
    smart_str,
    unicodify,
)
from galaxy.util.hash_util import hmac_new

encoding_sep = "__esep__"
encoding_sep2 = "__esepii__"


def tool_shed_decode(value):
    # Extract and verify hash
    value = unicodify(value)
    a, b = value.split(":")
    value = binascii.unhexlify(b)
    test = hmac_new(b"ToolShedAndGalaxyMustHaveThisSameKey", value)
    assert a == test
    # Restore from string
    values = None
    value = unicodify(value)
    try:
        values = json.loads(value)
    except Exception:
        pass
    if values is None:
        values = value
    return values


def tool_shed_encode(val):
    if isinstance(val, dict) or isinstance(val, list):
        value = json.dumps(val)
    else:
        value = val
    a = hmac_new(b"ToolShedAndGalaxyMustHaveThisSameKey", value)
    b = unicodify(binascii.hexlify(smart_str(value)))
    return f"{a}:{b}"



import functools
import os

try:
    import docutils.core
    import docutils.io
    import docutils.utils
    import docutils.writers.html4css1
except ImportError:
    docutils = None  # type: ignore[assignment]

from .custom_logging import get_logger


class FakeStream:
    def __init__(self, error):
        self.__error = error

    log_ = get_logger("docutils")

    def write(self, str):
        if len(str) > 0 and not str.isspace():
            if self.__error:
                raise Exception(str)
            self.log_.warning(str)


@functools.lru_cache(maxsize=None)
def get_publisher(error=False):
    docutils_writer = docutils.writers.html4css1.Writer()
    docutils_template_path = os.path.join(os.path.dirname(__file__), "docutils_template.txt")
    no_report_level = docutils.utils.Reporter.SEVERE_LEVEL + 1
    settings_overrides = {
        "embed_stylesheet": False,
        "template": docutils_template_path,
        "warning_stream": FakeStream(error),
        "doctitle_xform": False,  # without option, very different rendering depending on
        # number of sections in help content.
        "halt_level": no_report_level,
        "output_encoding": "unicode",
    }

    if not error:
        # in normal operation we don't want noisy warnings, that's tool author business
        settings_overrides["report_level"] = no_report_level

    Publisher = docutils.core.Publisher
    pub = Publisher(
        parser=None,
        writer=docutils_writer,
        settings=None,
        source_class=docutils.io.StringInput,
        destination_class=docutils.io.StringOutput,
    )
    pub.set_components("standalone", "restructuredtext", "pseudoxml")
    pub.process_programmatic_settings(None, settings_overrides, None)
    return pub


@functools.lru_cache(maxsize=None)
def rst_to_html(s, error=False):
    if docutils is None:
        raise Exception("Attempted to use rst_to_html but docutils unavailable.")

    publisher = get_publisher(error=error)
    publisher.set_source(s, None)
    publisher.set_destination(None, None)
    return publisher.publish(enable_exit_status=False)



"""Entry point for the usage of Cheetah templating within Galaxy."""

import traceback
from lib2to3.refactor import RefactoringTool

from Cheetah.Compiler import Compiler
from Cheetah.NameMapper import NotFound
from Cheetah.Parser import ParseError
from Cheetah.Template import Template
from packaging.version import Version
from past.translation import myfixes

from galaxy.util.tree_dict import TreeDict
from . import unicodify

# Skip libpasteurize fixers, which make sure code is py2 and py3 compatible.
# This is not needed, we only translate code on py3.
myfixes = [f for f in myfixes if not f.startswith("libpasteurize")]
refactoring_tool = RefactoringTool(myfixes, {"print_function": True})


class InputNotFoundSyntaxError(SyntaxError):
    pass


class FixedModuleCodeCompiler(Compiler):
    module_code = None

    def getModuleCode(self):
        self._moduleDef = self.module_code
        return self._moduleDef


def create_compiler_class(module_code):
    class CustomCompilerClass(FixedModuleCodeCompiler):
        pass

    CustomCompilerClass.module_code = module_code
    return CustomCompilerClass


def fill_template(
    template_text,
    context=None,
    retry=10,
    compiler_class=Compiler,
    first_exception=None,
    futurized=False,
    python_template_version="3",
    **kwargs,
):
    """Fill a cheetah template out for specified context.

    If template_text is None, an exception will be thrown, if context
    is None (the default) - keyword arguments to this function will be used
    as the context.
    """
    if template_text is None:
        raise TypeError("Template text specified as None to fill_template.")
    if not context:
        context = kwargs
    if isinstance(python_template_version, str):
        python_template_version = Version(python_template_version)
    try:
        klass = Template.compile(source=template_text, compilerClass=compiler_class)
    except ParseError as e:
        # Might happen on invalid syntax within a cheetah statement, like `#if $smxsize <> 128.0`
        if first_exception is None:
            first_exception = e
        if python_template_version.release[0] < 3 and retry > 0:
            module_code = Template.compile(
                source=template_text, compilerClass=compiler_class, returnAClass=False
            ).decode("utf-8")
            module_code = futurize_preprocessor(module_code)
            compiler_class = create_compiler_class(module_code)
            return fill_template(
                template_text=template_text,
                context=context,
                retry=retry - 1,
                compiler_class=compiler_class,
                first_exception=first_exception,
                python_template_version=python_template_version,
            )
        raise first_exception or e
    t = klass(searchList=[context])
    try:
        return unicodify(t, log_exception=False)
    except (NotFound, InputNotFoundSyntaxError) as e:
        if first_exception is None:
            first_exception = e
        if not isinstance(context, TreeDict):
            masked_input = None
            if "input" in context and callable(context["input"]):
                masked_input = context.pop("input", None)
            context = TreeDict(context)
            if "input" not in context and masked_input:
                context["input"] = masked_input
        tb = e.__traceback__
        if retry > 0:
            if python_template_version.release[0] < 3:
                last_stack = traceback.extract_tb(tb)[-1]
                if last_stack.name == "<listcomp>" and last_stack.lineno:
                    # On python 3 list, dict and set comprehensions as well as generator expressions
                    # have their own local scope, which prevents accessing frame variables in cheetah.
                    # We can work around this by replacing `$var` with `var`, but we only do this for
                    # list comprehensions, as this has never worked for dict or set comprehensions or
                    # generator expressions in Cheetah.
                    var_not_found = e.args[0].split("'")[1]
                    replace_str = f'VFFSL(SL,"{var_not_found}",True)'
                    lineno = last_stack.lineno - 1
                    module_code = t._CHEETAH_generatedModuleCode.splitlines()
                    module_code[lineno] = module_code[lineno].replace(replace_str, var_not_found)
                    module_code = "\n".join(module_code)
                    compiler_class = create_compiler_class(module_code)
            return fill_template(
                template_text=template_text,
                context=context,
                retry=retry - 1,
                compiler_class=compiler_class,
                first_exception=first_exception,
                python_template_version=python_template_version,
            )
        raise first_exception or e
    except Exception as e:
        if first_exception is None:
            first_exception = e
        if python_template_version.release[0] < 3 and not futurized:
            # Possibly an error caused by attempting to run python 2
            # template code on python 3. Run the generated module code
            # through futurize and hope for the best.
            module_code = t._CHEETAH_generatedModuleCode
            module_code = futurize_preprocessor(module_code)
            compiler_class = create_compiler_class(module_code)
            return fill_template(
                template_text=template_text,
                context=context,
                retry=retry,
                compiler_class=compiler_class,
                first_exception=first_exception,
                futurized=True,
                python_template_version=python_template_version,
            )
        raise first_exception or e


def futurize_preprocessor(source):
    source = str(refactoring_tool.refactor_string(source, name="auto_translate_cheetah"))
    # libfuturize.fixes.fix_unicode_keep_u' breaks from Cheetah.compat import unicode
    source = source.replace("from Cheetah.compat import str", "from Cheetah.compat import unicode")
    return source



"""Galaxy "safe" path functions forced to work with Windows-style paths regardless of current platform
"""

import ntpath
import sys

from . import _build_self

_build_self(sys.modules[__name__], ntpath)



"""
HTML Sanitizer (lists of acceptable_* ripped from feedparser)
"""
import bleach

from galaxy.util import unicodify

_acceptable_elements = [
    "a",
    "abbr",
    "acronym",
    "address",
    "area",
    "article",
    "aside",
    "audio",
    "b",
    "big",
    "blockquote",
    "br",
    "button",
    "canvas",
    "caption",
    "center",
    "cite",
    "code",
    "col",
    "colgroup",
    "command",
    "datagrid",
    "datalist",
    "dd",
    "del",
    "details",
    "dfn",
    "dialog",
    "dir",
    "div",
    "dl",
    "dt",
    "em",
    "event-source",
    "fieldset",
    "figure",
    "footer",
    "font",
    "form",
    "header",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "hr",
    "i",
    "img",
    "input",
    "ins",
    "keygen",
    "kbd",
    "label",
    "legend",
    "li",
    "m",
    "map",
    "menu",
    "meter",
    "multicol",
    "nav",
    "nextid",
    "ol",
    "output",
    "optgroup",
    "option",
    "p",
    "pre",
    "progress",
    "q",
    "s",
    "samp",
    "section",
    "select",
    "small",
    "sound",
    "source",
    "spacer",
    "span",
    "strike",
    "strong",
    "sub",
    "sup",
    "table",
    "tbody",
    "td",
    "textarea",
    "time",
    "tfoot",
    "th",
    "thead",
    "tr",
    "tt",
    "u",
    "ul",
    "var",
    "video",
    "noscript",
]

_acceptable_attributes = [
    "abbr",
    "accept",
    "accept-charset",
    "accesskey",
    "action",
    "align",
    "alt",
    "autocomplete",
    "autofocus",
    "axis",
    "background",
    "balance",
    "bgcolor",
    "bgproperties",
    "border",
    "bordercolor",
    "bordercolordark",
    "bordercolorlight",
    "bottompadding",
    "cellpadding",
    "cellspacing",
    "ch",
    "challenge",
    "char",
    "charoff",
    "choff",
    "charset",
    "checked",
    "cite",
    "class",
    "clear",
    "color",
    "cols",
    "colspan",
    "compact",
    "contenteditable",
    "controls",
    "coords",
    "data",
    "datafld",
    "datapagesize",
    "datasrc",
    "datetime",
    "default",
    "delay",
    "dir",
    "disabled",
    "draggable",
    "dynsrc",
    "enctype",
    "end",
    "face",
    "for",
    "form",
    "frame",
    "galleryimg",
    "gutter",
    "headers",
    "height",
    "hidefocus",
    "hidden",
    "high",
    "href",
    "hreflang",
    "hspace",
    "icon",
    "id",
    "inputmode",
    "ismap",
    "keytype",
    "label",
    "leftspacing",
    "lang",
    "list",
    "longdesc",
    "loop",
    "loopcount",
    "loopend",
    "loopstart",
    "low",
    "lowsrc",
    "max",
    "maxlength",
    "media",
    "method",
    "min",
    "multiple",
    "name",
    "nohref",
    "noshade",
    "nowrap",
    "open",
    "optimum",
    "pattern",
    "ping",
    "point-size",
    "prompt",
    "pqg",
    "radiogroup",
    "readonly",
    "rel",
    "repeat-max",
    "repeat-min",
    "replace",
    "required",
    "rev",
    "rightspacing",
    "rows",
    "rowspan",
    "rules",
    "scope",
    "selected",
    "shape",
    "size",
    "span",
    "src",
    "start",
    "step",
    "summary",
    "suppress",
    "tabindex",
    "target",
    "template",
    "title",
    "toppadding",
    "type",
    "unselectable",
    "usemap",
    "urn",
    "valign",
    "value",
    "variable",
    "volume",
    "vspace",
    "vrml",
    "width",
    "wrap",
    "xml:lang",
]


def sanitize_html(htmlSource, allow_data_urls=False):
    kwd = dict(tags=_acceptable_elements, attributes=_acceptable_attributes, strip=True)
    if allow_data_urls:
        kwd["protocols"] = list(bleach.ALLOWED_PROTOCOLS) + ["data"]
    return bleach.clean(unicodify(htmlSource), **kwd)



from pathlib import Path
from types import ModuleType
from typing import (
    Any,
    cast,
    Dict,
    Generator,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Type,
    Union,
)

import yaml

from galaxy.util import parse_xml
from galaxy.util.submodules import import_submodules

PathT = Union[str, Path]
PluginDictConfigT = Dict[str, Any]
PluginConfigsT = Union[PluginDictConfigT, List[PluginDictConfigT]]


class PluginConfigSource(NamedTuple):
    type: str
    source: Any


def plugins_dict(module: ModuleType, plugin_type_identifier: str) -> Dict[str, Type]:
    """Walk through all classes in submodules of module and find ones labelled
    with specified plugin_type_identifier and throw in a dictionary to allow
    constructions from plugins by these types later on.
    """
    plugin_dict = {}

    for plugin_module in import_submodules(module, ordered=True):
        for clazz in __plugin_classes_in_module(plugin_module):
            plugin_type = getattr(clazz, plugin_type_identifier, None)
            if plugin_type:
                plugin_dict[plugin_type] = clazz

    return plugin_dict


def load_plugins(
    plugins_dict: Dict[str, Type],
    plugin_source: PluginConfigSource,
    extra_kwds: Optional[Dict[str, Any]] = None,
    plugin_type_keys: Iterable[str] = ("type",),
    dict_to_list_key: Optional[str] = None,
) -> List[Any]:
    if extra_kwds is None:
        extra_kwds = {}
    if plugin_source.type == "xml":
        return __load_plugins_from_element(plugins_dict, plugin_source.source, extra_kwds)
    else:
        return __load_plugins_from_dicts(
            plugins_dict,
            plugin_source.source,
            extra_kwds,
            plugin_type_keys=plugin_type_keys,
            dict_to_list_key=dict_to_list_key,
        )


def __plugin_classes_in_module(plugin_module: ModuleType) -> Generator[Type, None, None]:
    for clazz in getattr(plugin_module, "__all__", []):
        try:
            clazz = getattr(plugin_module, clazz)
        except TypeError:
            clazz = clazz
        yield clazz


def __load_plugins_from_element(
    plugins_dict: Dict[str, Type], plugins_element, extra_kwds: Dict[str, Any]
) -> List[Any]:
    plugins = []

    for plugin_element in plugins_element:
        plugin_type = plugin_element.tag
        plugin_kwds = dict(plugin_element.items())
        plugin_kwds.update(extra_kwds)
        try:
            plugin_klazz = plugins_dict[plugin_type]
        except KeyError:
            template = "Failed to find plugin of type [%s] in available plugin types %s"
            message = template % (plugin_type, str(plugins_dict.keys()))
            raise Exception(message)

        plugin = plugin_klazz(**plugin_kwds)
        plugins.append(plugin)

    return plugins


def __load_plugins_from_dicts(
    plugins_dict: Dict[str, Type],
    configs: PluginConfigsT,
    extra_kwds: Dict[str, Any],
    plugin_type_keys: Iterable[str],
    dict_to_list_key: Optional[str],
) -> List[Any]:
    plugins = []

    configs_as_list: List[PluginDictConfigT]
    if isinstance(configs, dict) and dict_to_list_key is not None:
        configs_as_list = []
        for key, value in configs.items():
            config = value.copy()
            config[dict_to_list_key] = key
            configs_as_list.append(config)
    else:
        configs_as_list = cast(List[PluginDictConfigT], configs)

    for config in configs_as_list:
        plugin_type = None
        for plugin_type_key in plugin_type_keys:
            if plugin_type_key in config:
                plugin_type = config[plugin_type_key]
                break
        assert plugin_type is not None, f"Could not determine plugin type for [{config}]"
        plugin_kwds = config
        if extra_kwds:
            plugin_kwds = plugin_kwds.copy()
            plugin_kwds.update(extra_kwds)
        plugin = plugins_dict[plugin_type](**plugin_kwds)
        plugins.append(plugin)

    return plugins


def plugin_source_from_path(path: PathT) -> PluginConfigSource:
    filename = str(path)
    if (
        filename.endswith(".yaml")
        or filename.endswith(".yml")
        or filename.endswith(".yaml.sample")
        or filename.endswith(".yml.sample")
    ):
        return PluginConfigSource("dict", __read_yaml(path))
    else:
        return PluginConfigSource("xml", parse_xml(path, remove_comments=True).getroot())


def plugin_source_from_dict(as_dict: PluginConfigsT) -> PluginConfigSource:
    return PluginConfigSource("dict", as_dict)


def __read_yaml(path: PathT):
    if yaml is None:
        raise ImportError("Attempting to read YAML configuration file - but PyYAML dependency unavailable.")

    with open(path, "rb") as f:
        return yaml.safe_load(f)



import json
import logging
import os
from typing import (
    Optional,
    TYPE_CHECKING,
)
from urllib.parse import urljoin

from routes import url_for
from typing_extensions import Protocol

from galaxy import util
from galaxy.util.tool_shed import encoding_util

if TYPE_CHECKING:
    from .tool_shed_registry import Registry as ToolShedRegistry

log = logging.getLogger(__name__)

REPOSITORY_OWNER = "devteam"


class HasToolShedRegistry(Protocol):
    tool_shed_registry: "ToolShedRegistry"
    name: str


def accumulate_tool_dependencies(tool_shed_accessible, tool_dependencies, all_tool_dependencies):
    if tool_shed_accessible:
        if tool_dependencies:
            for tool_dependency in tool_dependencies:
                if tool_dependency not in all_tool_dependencies:
                    all_tool_dependencies.append(tool_dependency)
    return all_tool_dependencies


def check_tool_tag_set(elem, migrated_tool_configs_dict, missing_tool_configs_dict):
    file_path = elem.get("file", None)
    if file_path:
        name = os.path.basename(file_path)
        for migrated_tool_config in migrated_tool_configs_dict.keys():
            if migrated_tool_config in [file_path, name]:
                missing_tool_configs_dict[name] = migrated_tool_configs_dict[migrated_tool_config]
    return missing_tool_configs_dict


def generate_clone_url_for_installed_repository(app: HasToolShedRegistry, repository) -> str:
    """Generate the URL for cloning a repository that has been installed into a Galaxy instance."""
    tool_shed_url = get_tool_shed_url_from_tool_shed_registry(app, str(repository.tool_shed))
    return util.build_url(tool_shed_url, pathspec=["repos", str(repository.owner), str(repository.name)])


def generate_clone_url_for_repository_in_tool_shed(user, repository) -> str:
    """Generate the URL for cloning a repository that is in the tool shed."""
    base_url = url_for("/", qualified=True).rstrip("/")
    if user:
        protocol, base = base_url.split("://")
        username = f"{user.username}@"
        return f"{protocol}://{username}{base}/repos/{repository.user.username}/{repository.name}"
    else:
        return f"{base_url}/repos/{repository.user.username}/{repository.name}"


def generate_clone_url_from_repo_info_tup(app: HasToolShedRegistry, repo_info_tup) -> str:
    """Generate the URL for cloning a repository given a tuple of toolshed, name, owner, changeset_revision."""
    # Example tuple: ['http://localhost:9009', 'blast_datatypes', 'test', '461a4216e8ab', False]
    (
        toolshed,
        name,
        owner,
        changeset_revision,
        prior_installation_required,
        only_if_compiling_contained_td,
    ) = parse_repository_dependency_tuple(repo_info_tup)
    tool_shed_url = get_tool_shed_url_from_tool_shed_registry(app, toolshed)
    # Don't include the changeset_revision in clone urls.
    return util.build_url(tool_shed_url, pathspec=["repos", owner, name])


def get_repository_dependencies(app, tool_shed_url, repository_name, repository_owner, changeset_revision):
    repository_dependencies_dict = {}
    tool_shed_accessible = True
    params = dict(name=repository_name, owner=repository_owner, changeset_revision=changeset_revision)
    pathspec = ["repository", "get_repository_dependencies"]
    try:
        raw_text = util.url_get(
            tool_shed_url, auth=app.tool_shed_registry.url_auth(tool_shed_url), pathspec=pathspec, params=params
        )
        tool_shed_accessible = True
    except Exception as e:
        tool_shed_accessible = False
        log.warning(
            "The URL\n%s\nraised the exception:\n%s\n",
            util.build_url(tool_shed_url, pathspec=pathspec, params=params),
            e,
        )
    if tool_shed_accessible:
        if len(raw_text) > 2:
            encoded_text = json.loads(util.unicodify(raw_text))
            repository_dependencies_dict = encoding_util.tool_shed_decode(encoded_text)
    return tool_shed_accessible, repository_dependencies_dict


def get_protocol_from_tool_shed_url(tool_shed_url: str) -> str:
    """Return the protocol from the received tool_shed_url if it exists."""
    try:
        if tool_shed_url.find("://") > 0:
            return tool_shed_url.split("://")[0].lower()
    except Exception:
        # We receive a lot of calls here where the tool_shed_url is None.  The container_util uses
        # that value when creating a header row.  If the tool_shed_url is not None, we have a problem.
        if tool_shed_url is not None:
            log.exception("Handled exception getting the protocol from Tool Shed URL %s", str(tool_shed_url))
    # Default to HTTP protocol.
    return "http"


def get_tool_shed_repository_ids(as_string=False, **kwd):
    tsrid = kwd.get("tool_shed_repository_id", None)
    tsridslist = util.listify(kwd.get("tool_shed_repository_ids", None))
    if not tsridslist:
        tsridslist = util.listify(kwd.get("id", None))
    if tsridslist is not None:
        if tsrid is not None and tsrid not in tsridslist:
            tsridslist.append(tsrid)
        if as_string:
            return ",".join(tsridslist)
        return tsridslist
    else:
        tsridslist = util.listify(kwd.get("ordered_tsr_ids", None))
        if tsridslist is not None:
            if as_string:
                return ",".join(tsridslist)
            return tsridslist
    if as_string:
        return ""
    return []


def get_tool_shed_url_from_tool_shed_registry(app: HasToolShedRegistry, tool_shed: str) -> Optional[str]:
    """
    The value of tool_shed is something like: toolshed.g2.bx.psu.edu.  We need the URL to this tool shed, which is
    something like: http://toolshed.g2.bx.psu.edu/
    """
    cleaned_tool_shed = remove_protocol_from_tool_shed_url(tool_shed)
    for shed_url in app.tool_shed_registry.tool_sheds.values():
        if shed_url.find(cleaned_tool_shed) >= 0:
            if shed_url.endswith("/"):
                shed_url = shed_url.rstrip("/")
            return shed_url
    # The tool shed from which the repository was originally installed must no longer be configured in tool_sheds_conf.xml.
    return None


def get_tool_shed_repository_url(app: HasToolShedRegistry, tool_shed: str, owner: str, name: str):
    tool_shed_url = get_tool_shed_url_from_tool_shed_registry(app, tool_shed)
    if tool_shed_url:
        # Append a slash to the tool shed URL, because urlparse.urljoin will eliminate
        # the last part of a URL if it does not end with a forward slash.
        tool_shed_url = f"{tool_shed_url}/"
        return urljoin(tool_shed_url, f"view/{owner}/{name}")
    return tool_shed_url


def get_user_by_username(app, username):
    """Get a user from the database by username."""
    sa_session = app.model.session
    try:
        user = sa_session.query(app.model.User).filter(app.model.User.table.c.username == username).one()
        return user
    except Exception:
        return None


def handle_galaxy_url(trans, **kwd):
    galaxy_url = kwd.get("galaxy_url", None)
    if galaxy_url:
        trans.set_cookie(galaxy_url, name="toolshedgalaxyurl")
    else:
        galaxy_url = trans.get_cookie(name="toolshedgalaxyurl")
    return galaxy_url


def handle_tool_shed_url_protocol(app: HasToolShedRegistry, shed_url: str) -> str:
    """Handle secure and insecure HTTP protocol since they may change over time."""
    try:
        if app.name == "galaxy":
            url = remove_protocol_from_tool_shed_url(shed_url)
            tool_shed_url = get_tool_shed_url_from_tool_shed_registry(app, url)
            assert tool_shed_url
        else:
            tool_shed_url = str(url_for("/", qualified=True)).rstrip("/")
        return tool_shed_url
    except Exception:
        # We receive a lot of calls here where the tool_shed_url is None.  The container_util uses
        # that value when creating a header row.  If the tool_shed_url is not None, we have a problem.
        if shed_url is not None:
            log.exception("Handled exception removing protocol from URL %s", str(shed_url))
        return shed_url


def parse_repository_dependency_tuple(repository_dependency_tuple, contains_error=False):
    # Default both prior_installation_required and only_if_compiling_contained_td to False in cases where metadata should be reset on the
    # repository containing the repository_dependency definition.
    prior_installation_required = "False"
    only_if_compiling_contained_td = "False"
    if contains_error:
        if len(repository_dependency_tuple) == 5:
            tool_shed, name, owner, changeset_revision, error = repository_dependency_tuple
        elif len(repository_dependency_tuple) == 6:
            tool_shed, name, owner, changeset_revision, prior_installation_required, error = repository_dependency_tuple
        elif len(repository_dependency_tuple) == 7:
            (
                tool_shed,
                name,
                owner,
                changeset_revision,
                prior_installation_required,
                only_if_compiling_contained_td,
                error,
            ) = repository_dependency_tuple
        return (
            tool_shed,
            name,
            owner,
            changeset_revision,
            prior_installation_required,
            only_if_compiling_contained_td,
            error,
        )
    else:
        if len(repository_dependency_tuple) == 4:
            tool_shed, name, owner, changeset_revision = repository_dependency_tuple
        elif len(repository_dependency_tuple) == 5:
            tool_shed, name, owner, changeset_revision, prior_installation_required = repository_dependency_tuple
        elif len(repository_dependency_tuple) == 6:
            (
                tool_shed,
                name,
                owner,
                changeset_revision,
                prior_installation_required,
                only_if_compiling_contained_td,
            ) = repository_dependency_tuple
        return tool_shed, name, owner, changeset_revision, prior_installation_required, only_if_compiling_contained_td


def remove_port_from_tool_shed_url(tool_shed_url: str) -> str:
    """Return a partial Tool Shed URL, eliminating the port if it exists."""
    try:
        if tool_shed_url.find(":") > 0:
            # Eliminate the port, if any, since it will result in an invalid directory name.
            new_tool_shed_url = tool_shed_url.split(":")[0]
        else:
            new_tool_shed_url = tool_shed_url
        return new_tool_shed_url.rstrip("/")
    except Exception:
        # We receive a lot of calls here where the tool_shed_url is None.  The container_util uses
        # that value when creating a header row.  If the tool_shed_url is not None, we have a problem.
        if tool_shed_url is not None:
            log.exception("Handled exception removing the port from Tool Shed URL %s", str(tool_shed_url))
        return tool_shed_url


def remove_protocol_and_port_from_tool_shed_url(tool_shed_url: str) -> str:
    """Return a partial Tool Shed URL, eliminating the protocol and/or port if either exists."""
    tool_shed = remove_protocol_from_tool_shed_url(tool_shed_url)
    tool_shed = remove_port_from_tool_shed_url(tool_shed)
    return tool_shed


def remove_protocol_and_user_from_clone_url(repository_clone_url: str) -> str:
    """Return a URL that can be used to clone a repository, eliminating the protocol and user if either exists."""
    if repository_clone_url.find("@") > 0:
        # We have an url that includes an authenticated user, something like:
        # http://test@bx.psu.edu:9009/repos/some_username/column
        items = repository_clone_url.split("@")
        tmp_url = items[1]
    elif repository_clone_url.find("//") > 0:
        # We have an url that includes only a protocol, something like:
        # http://bx.psu.edu:9009/repos/some_username/column
        items = repository_clone_url.split("//")
        tmp_url = items[1]
    else:
        tmp_url = repository_clone_url
    return tmp_url.rstrip("/")


def remove_protocol_from_tool_shed_url(tool_shed_url: str) -> str:
    """Return a partial Tool Shed URL, eliminating the protocol if it exists."""
    return util.remove_protocol_from_url(tool_shed_url)


__all__ = (
    "accumulate_tool_dependencies",
    "check_tool_tag_set",
    "generate_clone_url_for_installed_repository",
    "generate_clone_url_for_repository_in_tool_shed",
    "generate_clone_url_from_repo_info_tup",
    "get_repository_dependencies",
    "get_protocol_from_tool_shed_url",
    "get_tool_shed_repository_ids",
    "get_tool_shed_url_from_tool_shed_registry",
    "get_tool_shed_repository_url",
    "get_user_by_username",
    "handle_galaxy_url",
    "handle_tool_shed_url_protocol",
    "parse_repository_dependency_tuple",
    "remove_port_from_tool_shed_url",
    "remove_protocol_and_port_from_tool_shed_url",
    "remove_protocol_and_user_from_clone_url",
    "remove_protocol_from_tool_shed_url",
)



from .dynamic import HasDynamicProperties


class Bunch(HasDynamicProperties):
    """
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52308

    Often we want to just collect a bunch of stuff together, naming each item of
    the bunch; a dictionary's OK for that, but a small do-nothing class is even handier, and prettier to use.
    """

    def __init__(self, **kwds):
        self.__dict__.update(kwds)

    def dict(self):
        return self.__dict__

    def get(self, key, default=None):
        return self.__dict__.get(key, default)

    def __iter__(self):
        return iter(self.__dict__)

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def __str__(self):
        return f"{self.__dict__}"

    def __bool__(self):
        return bool(self.__dict__)

    __nonzero__ = __bool__

    def __setitem__(self, k, v):
        self.__dict__.__setitem__(k, v)

    def __contains__(self, item):
        return item in self.__dict__



import datetime
import uuid
from typing import (
    Any,
    Callable,
    Dict,
    Optional,
)


def dict_for(obj, **kwds):
    # Create dict to represent item.
    return dict(model_class=obj.__class__.__name__, **kwds)


class Dictifiable:
    """Mixin that enables objects to be converted to dictionaries. This is useful
    when for sharing objects across boundaries, such as the API, tool scripts,
    and JavaScript code."""

    def to_dict(self, view: str = "collection", value_mapper: Optional[Dict[str, Callable]] = None) -> Dict[str, Any]:
        """
        Return item dictionary.
        """

        if not value_mapper:
            value_mapper = {}

        def get_value(key, item):
            """
            Recursive helper function to get item values.
            """
            # FIXME: why use exception here? Why not look for key in value_mapper
            # first and then default to to_dict?
            try:
                return item.to_dict(view=view, value_mapper=value_mapper)
            except Exception:
                assert value_mapper is not None
                if key in value_mapper:
                    return value_mapper[key](item)
                if type(item) == datetime.datetime:
                    return item.isoformat()
                elif type(item) == uuid.UUID:
                    return str(item)
                # Leaving this for future reference, though we may want a more
                # generic way to handle special type mappings going forward.
                # If the item is of a class that needs to be 'stringified' before being put into a JSON data structure
                # elif type(item) in []:
                #    return str(item)
                return item

        # Create dict to represent item.
        rval = dict_for(self)

        # Fill item dict with visible keys.
        try:
            visible_keys = self.__getattribute__(f"dict_{view}_visible_keys")
        except AttributeError:
            raise Exception(f"Unknown Dictifiable view: {view}")
        for key in visible_keys:
            try:
                item = self.__getattribute__(key)
                if isinstance(item, list):
                    rval[key] = []
                    for i in item:
                        rval[key].append(get_value(key, i))
                else:
                    rval[key] = get_value(key, item)
            except AttributeError:
                rval[key] = None

        return rval



import bz2
import gzip
import os
import re
import tarfile
import zipfile
from io import (
    BytesIO,
    StringIO,
)
from typing import (
    Dict,
    IO,
    Tuple,
)

from typing_extensions import Protocol

from galaxy import util
from galaxy.util.image_util import image_type

HTML_CHECK_LINES = 100
CHUNK_SIZE = 2**15  # 32Kb
HTML_REGEXPS = (
    re.compile(r"<A\s+[^>]*HREF[^>]+>", re.I),
    re.compile(r"<IFRAME[^>]*>", re.I),
    re.compile(r"<FRAMESET[^>]*>", re.I),
    re.compile(r"<META[\W][^>]*>", re.I),
    re.compile(r"<SCRIPT[^>]*>", re.I),
)


class CompressionChecker(Protocol):
    def __call__(self, file_path: str, check_content: bool = True) -> Tuple[bool, bool]:
        ...


def check_html(name, file_path: bool = True) -> bool:
    """
    Returns True if the file/string contains HTML code.
    """
    # Handles files if file_path is True or text if file_path is False
    temp: IO[str]
    if file_path:
        temp = open(name, encoding="utf-8")
    else:
        temp = StringIO(util.unicodify(name))
    try:
        for _ in range(HTML_CHECK_LINES):
            line = temp.readline(CHUNK_SIZE)
            if not line:
                break
            if any(regexp.search(line) for regexp in HTML_REGEXPS):
                return True
    except UnicodeDecodeError:
        return False
    finally:
        temp.close()
    return False


def check_binary(name, file_path: bool = True) -> bool:
    # Handles files if file_path is True or text if file_path is False
    temp: IO[bytes]
    if file_path:
        temp = open(name, "rb")
        size = os.stat(name).st_size
    else:
        temp = BytesIO(name)
        size = len(name)
    read_start = int(size / 2)
    read_length = 1024
    try:
        if util.is_binary(temp.read(read_length)):
            return True
        # Some binary files have text only within the first 1024
        # Read 1024 from the middle of the file if this is not
        # a gzip or zip compressed file (bzip are indexed),
        # to avoid issues with long txt headers on binary files.
        if file_path and not is_gzip(name) and not is_zip(name) and not is_bz2(name):
            # file_path=False doesn't seem to be used in the codebase
            temp.seek(read_start)
            return util.is_binary(temp.read(read_length))
        return False
    finally:
        temp.close()


def check_gzip(file_path: str, check_content: bool = True) -> Tuple[bool, bool]:
    # This method returns a tuple of booleans representing ( is_gzipped, is_valid )
    # Make sure we have a gzipped file
    try:
        with open(file_path, "rb") as temp:
            magic_check = temp.read(2)
        if magic_check != util.gzip_magic:
            return (False, False)
    except Exception:
        return (False, False)
    # We support some binary data types, so check if the compressed binary file is valid
    # If the file is Bam, it should already have been detected as such, so we'll just check
    # for sff format.
    try:
        with gzip.open(file_path, "rb") as fh:
            header = fh.read(4)
        if header == b".sff":
            return (True, True)
    except Exception:
        return (False, False)

    if not check_content:
        return (True, True)

    with gzip.open(file_path, mode="rb") as gzipped_file:
        chunk = gzipped_file.read(CHUNK_SIZE)
    # See if we have a compressed HTML file
    if check_html(chunk, file_path=False):
        return (True, False)
    return (True, True)


def check_bz2(file_path: str, check_content: bool = True) -> Tuple[bool, bool]:
    try:
        with open(file_path, "rb") as temp:
            magic_check = temp.read(3)
        if magic_check != util.bz2_magic:
            return (False, False)
    except Exception:
        return (False, False)

    if not check_content:
        return (True, True)

    with bz2.BZ2File(file_path, mode="rb") as bzipped_file:
        chunk = bzipped_file.read(CHUNK_SIZE)
    # See if we have a compressed HTML file
    if check_html(chunk, file_path=False):
        return (True, False)
    return (True, True)


def check_zip(file_path: str, check_content: bool = True, files=1) -> Tuple[bool, bool]:
    if not zipfile.is_zipfile(file_path):
        return (False, False)

    if not check_content:
        return (True, True)

    chunk = None
    for filect, member in enumerate(iter_zip(file_path)):
        handle, name = member
        chunk = handle.read(CHUNK_SIZE)
        if chunk and check_html(chunk, file_path=False):
            return (True, False)
        if filect >= files:
            break
    return (True, True)


def is_bz2(file_path: str) -> bool:
    is_bz2, is_valid = check_bz2(file_path, check_content=False)
    return is_bz2


def is_gzip(file_path: str) -> bool:
    is_gzipped, is_valid = check_gzip(file_path, check_content=False)
    return is_gzipped


def is_zip(file_path: str) -> bool:
    is_zipped, is_valid = check_zip(file_path, check_content=False)
    return is_zipped


def is_single_file_zip(file_path: str) -> bool:
    for i, _ in enumerate(iter_zip(file_path)):
        if i > 1:
            return False
    return True


def is_tar(file_path: str) -> bool:
    return tarfile.is_tarfile(file_path)


def iter_zip(file_path: str):
    with zipfile.ZipFile(file_path) as z:
        for f in filter(lambda x: not x.endswith("/"), z.namelist()):
            yield (z.open(f), f)


def check_image(file_path: str):
    """Simple wrapper around image_type to yield a True/False verdict"""
    if image_type(file_path):
        return True
    return False


COMPRESSION_CHECK_FUNCTIONS: Dict[str, CompressionChecker] = {
    "gzip": check_gzip,
    "bz2": check_bz2,
    "zip": check_zip,
}


__all__ = (
    "check_binary",
    "check_bz2",
    "check_gzip",
    "check_html",
    "check_image",
    "check_zip",
    "COMPRESSION_CHECK_FUNCTIONS",
    "is_gzip",
    "is_bz2",
    "is_zip",
)



import bz2
import gzip
import io
import logging
import os
import tarfile
import tempfile
import zipfile
from typing import (
    Any,
    cast,
    Generator,
    IO,
    Iterable,
    List,
    Optional,
    overload,
    Tuple,
    Union,
)

from typing_extensions import Literal

from galaxy.util.path import (
    safe_relpath,
    StrPath,
)
from .checkers import (
    is_bz2,
    is_gzip,
)

log = logging.getLogger(__name__)

FileObjTypeStr = Union[IO[str], io.TextIOWrapper]
FileObjTypeBytes = Union[gzip.GzipFile, bz2.BZ2File, IO[bytes]]
FileObjType = Union[FileObjTypeStr, FileObjTypeBytes]


@overload
def get_fileobj(filename: str, mode: Literal["r"], compressed_formats: Optional[List[str]] = None) -> FileObjTypeStr:
    ...


@overload
def get_fileobj(filename: str, mode: Literal["rb"], compressed_formats: Optional[List[str]] = None) -> FileObjTypeBytes:
    ...


@overload
def get_fileobj(filename: str) -> FileObjTypeStr:
    ...


@overload
def get_fileobj(filename: str, mode: str = "r", compressed_formats: Optional[List[str]] = None) -> FileObjType:
    ...


def get_fileobj(filename: str, mode: str = "r", compressed_formats: Optional[List[str]] = None) -> FileObjType:
    """
    Returns a fileobj. If the file is compressed, return an appropriate file
    reader. In text mode, always use 'utf-8' encoding.

    :param filename: path to file that should be opened
    :param mode: mode to pass to opener
    :param compressed_formats: list of allowed compressed file formats among
      'bz2', 'gzip' and 'zip'. If left to None, all 3 formats are allowed
    """
    return get_fileobj_raw(filename, mode, compressed_formats)[1]


@overload
def get_fileobj_raw(
    filename: str, mode: Literal["r"], compressed_formats: Optional[List[str]] = None
) -> Tuple[Optional[str], FileObjTypeStr]:
    ...


@overload
def get_fileobj_raw(
    filename: str, mode: Literal["rb"], compressed_formats: Optional[List[str]] = None
) -> Tuple[Optional[str], FileObjTypeBytes]:
    ...


@overload
def get_fileobj_raw(filename: str) -> Tuple[Optional[str], FileObjTypeStr]:
    ...


@overload
def get_fileobj_raw(
    filename: str, mode: str = "r", compressed_formats: Optional[List[str]] = None
) -> Tuple[Optional[str], FileObjType]:
    ...


def get_fileobj_raw(
    filename: str, mode: str = "r", compressed_formats: Optional[List[str]] = None
) -> Tuple[Optional[str], FileObjType]:
    if compressed_formats is None:
        compressed_formats = ["bz2", "gzip", "zip"]
    # Remove 't' from mode, which may cause an error for compressed files
    mode = mode.replace("t", "")
    # 'U' mode is deprecated, we open in 'r'.
    if mode == "U":
        mode = "r"
    compressed_format = None
    if "gzip" in compressed_formats and is_gzip(filename):
        fh: Union[gzip.GzipFile, bz2.BZ2File, IO[bytes]] = gzip.GzipFile(filename, mode)
        compressed_format = "gzip"
    elif "bz2" in compressed_formats and is_bz2(filename):
        mode = cast(Literal["a", "ab", "r", "rb", "w", "wb", "x", "xb"], mode)
        fh = bz2.BZ2File(filename, mode)
        compressed_format = "bz2"
    elif "zip" in compressed_formats and zipfile.is_zipfile(filename):
        # Return fileobj for the first file in a zip file.
        # 'b' is not allowed in the ZipFile mode argument
        # since it always opens files in binary mode.
        # For emulating text mode, we will be returning the binary fh in a
        # TextIOWrapper.
        zf_mode = cast(Literal["r", "w"], mode.replace("b", ""))
        with zipfile.ZipFile(filename, zf_mode) as zh:
            fh = zh.open(zh.namelist()[0], zf_mode)
        compressed_format = "zip"
    elif "b" in mode:
        return compressed_format, open(filename, mode)
    else:
        return compressed_format, open(filename, mode, encoding="utf-8")
    if "b" not in mode:
        return compressed_format, io.TextIOWrapper(cast(IO[bytes], fh), encoding="utf-8")
    else:
        return compressed_format, fh


def file_iter(fname: str, sep: Optional[Any] = None) -> Generator[Union[List[bytes], Any, List[str]], None, None]:
    """
    This generator iterates over a file and yields its lines
    splitted via the C{sep} parameter. Skips empty lines and lines starting with
    the C{#} character.

    >>> lines = [ line for line in file_iter(__file__) ]
    >>> len(lines) !=  0
    True
    """
    with get_fileobj(fname) as fh:
        for line in fh:
            if line and line[0] != "#":
                yield line.split(sep)


ArchiveMemberType = Union[tarfile.TarInfo, zipfile.ZipInfo]


def decompress_bytes_to_directory(content: bytes) -> str:
    temp_directory = tempfile.mkdtemp()
    with tempfile.NamedTemporaryFile(delete=False) as fp:
        fp.write(content)
        fp.close()
        return CompressedFile(fp.name).extract(temp_directory)


def decompress_path_to_directory(path: str) -> str:
    temp_directory = tempfile.mkdtemp()
    return CompressedFile(path).extract(temp_directory)


class CompressedFile:
    archive: Union[tarfile.TarFile, zipfile.ZipFile]

    @staticmethod
    def can_decompress(file_path: StrPath) -> bool:
        return tarfile.is_tarfile(file_path) or zipfile.is_zipfile(file_path)

    def __init__(self, file_path: StrPath, mode: str = "r") -> None:
        file_path_str = str(file_path)
        if tarfile.is_tarfile(file_path):
            self.file_type = "tar"
        elif zipfile.is_zipfile(file_path) and not file_path_str.endswith(".jar"):
            self.file_type = "zip"
        self.file_name = os.path.splitext(os.path.basename(file_path))[0]
        if self.file_name.endswith(".tar"):
            self.file_name = os.path.splitext(self.file_name)[0]
        self.type = self.file_type
        method = f"open_{self.file_type}"
        if hasattr(self, method):
            self.archive = getattr(self, method)(file_path, mode)
        else:
            raise NameError(f"File type {self.file_type} specified, no open method found.")

    @property
    def common_prefix_dir(self) -> str:
        """
        Get the common prefix directory for all the files in the archive, if any.

        Returns '' if the archive contains multiple files and/or directories at
        the root of the archive.
        """
        contents = self.getmembers()
        common_prefix = ""
        if len(contents) > 1:
            common_prefix = os.path.commonprefix([self.getname(item) for item in contents])
            # If the common_prefix does not end with a slash, check that is a
            # directory and all other files are contained in it
            common_prefix_member = self.getmember(common_prefix)
            if (
                len(common_prefix) >= 1
                and not common_prefix.endswith(os.sep)
                and common_prefix_member
                and self.isdir(common_prefix_member)
                and all(self.getname(item).startswith(common_prefix + os.sep) for item in contents if self.isfile(item))
            ):
                common_prefix += os.sep
            if not common_prefix.endswith(os.sep):
                common_prefix = ""
        return common_prefix

    def extract(self, path: StrPath) -> str:
        """Determine the path to which the archive should be extracted."""
        contents = self.getmembers()
        extraction_path = path
        common_prefix_dir = self.common_prefix_dir
        if len(contents) == 1:
            # The archive contains a single file, return the extraction path.
            if self.isfile(contents[0]):
                extraction_path = os.path.join(path, self.file_name)
                if not os.path.exists(extraction_path):
                    os.makedirs(extraction_path)
                if isinstance(self.archive, tarfile.TarFile):
                    members_t = cast(Iterable[tarfile.TarInfo], self.safemembers())
                    self.archive.extractall(extraction_path, members=members_t)
                else:
                    members_z = cast(Iterable[str], self.safemembers())
                    self.archive.extractall(extraction_path, members=members_z)
        else:
            if not common_prefix_dir:
                extraction_path = os.path.join(path, self.file_name)
                if not os.path.exists(extraction_path):
                    os.makedirs(extraction_path)
            if isinstance(self.archive, tarfile.TarFile):
                members_t = cast(Iterable[tarfile.TarInfo], self.safemembers())
                self.archive.extractall(extraction_path, members=members_t)
            else:
                members_z = cast(Iterable[str], self.safemembers())
                self.archive.extractall(extraction_path, members=members_z)
        # Since .zip files store unix permissions separately, we need to iterate through the zip file
        # and set permissions on extracted members.
        if self.file_type == "zip":
            assert isinstance(self.archive, zipfile.ZipFile)
            for zipped_file in contents:
                filename = self.getname(zipped_file)
                absolute_filepath = os.path.join(extraction_path, filename)
                external_attributes = self.archive.getinfo(filename).external_attr
                # The 2 least significant bytes are irrelevant, the next two contain unix permissions.
                unix_permissions = external_attributes >> 16
                if unix_permissions != 0:
                    if os.path.exists(absolute_filepath):
                        os.chmod(absolute_filepath, unix_permissions)
                    else:
                        log.warning(
                            f"Unable to change permission on extracted file '{absolute_filepath}' as it does not exist"
                        )
        return os.path.abspath(os.path.join(extraction_path, common_prefix_dir))

    def safemembers(self) -> Union[Iterable[tarfile.TarInfo], Iterable[str]]:
        members = self.archive
        common_prefix_dir = self.common_prefix_dir
        if self.file_type == "tar":
            assert isinstance(members, tarfile.TarFile)
            for finfo in members:
                if not safe_relpath(finfo.name):
                    raise Exception(f"Path '{finfo.name}' is blocked (illegal path).")
                if finfo.issym() or finfo.islnk():
                    link_target = os.path.join(os.path.dirname(finfo.name), finfo.linkname)
                    if not safe_relpath(link_target) or not os.path.normpath(link_target).startswith(common_prefix_dir):
                        raise Exception(f"Link '{finfo.name}' to '{finfo.linkname}' is blocked.")
                yield finfo
        elif self.file_type == "zip":
            assert isinstance(members, zipfile.ZipFile)
            for name in members.namelist():
                if not safe_relpath(name):
                    raise Exception(f"{name} is blocked (illegal path).")
                yield name

    def getmembers_tar(self) -> List[tarfile.TarInfo]:
        assert isinstance(self.archive, tarfile.TarFile)
        return self.archive.getmembers()

    def getmembers_zip(self) -> List[zipfile.ZipInfo]:
        assert isinstance(self.archive, zipfile.ZipFile)
        return self.archive.infolist()

    def getname_tar(self, item: tarfile.TarInfo) -> str:
        return item.name

    def getname_zip(self, item: zipfile.ZipInfo) -> str:
        return item.filename

    def getmember(self, name: str) -> Optional[ArchiveMemberType]:
        for member in self.getmembers():
            if self.getname(member) == name:
                return member
        return None

    def getmembers(self) -> List[ArchiveMemberType]:
        return cast(List[ArchiveMemberType], getattr(self, f"getmembers_{self.type}")())

    def getname(self, member: ArchiveMemberType) -> str:
        return cast(str, getattr(self, f"getname_{self.type}")(member))

    def isdir(self, member: ArchiveMemberType) -> bool:
        return cast(bool, getattr(self, f"isdir_{self.type}")(member))

    def isdir_tar(self, member: tarfile.TarInfo) -> bool:
        return member.isdir()

    def isdir_zip(self, member: zipfile.ZipInfo) -> bool:
        if member.filename.endswith(os.sep):
            return True
        return False

    def isfile(self, member: ArchiveMemberType) -> bool:
        if not self.isdir(member):
            return True
        return False

    def open_tar(self, filepath: StrPath, mode: str) -> tarfile.TarFile:
        return tarfile.open(filepath, mode, errorlevel=0)

    def open_zip(self, filepath: StrPath, mode: str) -> zipfile.ZipFile:
        mode = cast(Literal["a", "r", "w", "x"], mode)
        return zipfile.ZipFile(filepath, mode)

    def zipfile_ok(self, path_to_archive: StrPath) -> bool:
        """
        This function is a bit pedantic and not functionally necessary.  It checks whether there is
        no file pointing outside of the extraction, because ZipFile.extractall() has some potential
        security holes.  See python zipfile documentation for more details.
        """
        basename = os.path.realpath(os.path.dirname(path_to_archive))
        zip_archive = zipfile.ZipFile(path_to_archive)
        for member in zip_archive.namelist():
            member_path = os.path.realpath(os.path.join(basename, member))
            if not member_path.startswith(basename):
                return False
        return True



import re
import sqlite3

try:
    import sqlparse

    def is_read_only_query(query):
        statements = sqlparse.parse(query)
        for statement in statements:
            if statement.get_type() != "SELECT":
                return False
        return True

except ImportError:
    # Without sqlparse we use a very weak regex check
    def is_read_only_query(query):
        if re.match("select ", query, re.IGNORECASE):
            if re.search('^([^"]|"[^"]*")*?;', query) or re.search("^([^']|'[^']*')*?;", query):
                return False
            else:
                return True
        return False


def connect(path):
    connection = sqlite3.connect(path)
    connection.row_factory = sqlite3.Row
    return connection



import os
import zlib
from typing import (
    Dict,
    Iterator,
    List,
    Optional,
    Set,
)
from urllib.parse import quote

import zipstream

from .path import safe_walk

CRC32_MIN = 1444
CRC32_MAX = 1459


class ZipstreamWrapper:
    def __init__(
        self, archive_name: Optional[str] = None, upstream_mod_zip: bool = False, upstream_gzip: bool = False
    ) -> None:
        self.upstream_mod_zip = upstream_mod_zip
        self.archive_name = archive_name
        if not self.upstream_mod_zip:
            self.archive = zipstream.ZipFile(
                allowZip64=True, compression=zipstream.ZIP_STORED if upstream_gzip else zipstream.ZIP_DEFLATED
            )
        self.files: List[str] = []
        self.directories: Set[str] = set()
        self.size = 0

    def response(self) -> Iterator[bytes]:
        if self.upstream_mod_zip:
            dir_lines = [f"0 0 @directory {directory}" for directory in self.directories]
            yield "\n".join(dir_lines + self.files).encode()
        else:
            yield from iter(self.archive)

    def get_headers(self) -> Dict[str, str]:
        headers = {}
        if self.archive_name:
            headers["Content-Disposition"] = f'attachment; filename="{self.archive_name}.zip"'
        if self.upstream_mod_zip:
            headers["X-Archive-Files"] = "zip"
        else:
            headers["Content-Type"] = "application/x-zip-compressed"
        return headers

    def add_path(self, path: str, archive_name: str) -> None:
        size = int(os.stat(path).st_size)
        if self.upstream_mod_zip:
            # calculating crc32 would defeat the point of using mod-zip, but if we ever calculate hashsums we should consider this
            crc32 = "-"
            # We do have to calculate the crc32 for files that are between 1444 and 1459 bytes in size, xref: https://github.com/evanmiller/mod_zip/issues/44#issuecomment-656660686
            # Oddly that seems to be only true for usegalaxy.org (nginx version 1.12.2), and works fine locally (nginx 1.19.10).
            # May have been fixed in nginx 1.17.0
            if CRC32_MIN <= os.path.getsize(path) <= CRC32_MAX:
                with open(path, "rb") as contents:
                    crc32 = hex(zlib.crc32(contents.read()))[2:]
            line = f"{crc32} {size} {quote(path)} {archive_name}"
            head, tail = os.path.split(archive_name)
            if head:
                self.directories.add(head)
            self.files.append(line)
        else:
            self.size += size
            self.archive.write(path, archive_name)

    def write(self, path: str, archive_name: Optional[str] = None) -> None:
        if os.path.isdir(path):
            pardir = os.path.join(path, os.pardir)
            for root, directories, files in safe_walk(path):
                for directory in directories:
                    dir_path = os.path.join(root, directory)
                    self.add_path(dir_path, os.path.relpath(dir_path, pardir))
                for file in files:
                    file_path = os.path.join(root, file)
                    self.add_path(file_path, os.path.relpath(file_path, pardir))
        else:
            self.add_path(path, archive_name or os.path.basename(path))



from typing import (
    Dict,
    Union,
)

Theme = Dict[str, Union["Theme", str]]


def flatten_theme(theme: Theme, prefix: str = "-") -> Dict[str, str]:
    """Transforms a nested theme dictionary into a flat dictionary,
    containing keys compatible with css variables. e.g. '--masthead-background-color'"""
    flat_attributes: Dict[str, str] = {}

    for key, val in theme.items():
        if isinstance(val, str):
            flat_attributes[f"{prefix}-{key}"] = val
        elif isinstance(val, Dict):
            flat_attributes.update(flatten_theme(val, f"{prefix}-{key}"))

    return flat_attributes



import logging
from typing import (
    Dict,
    NamedTuple,
    Optional,
)

from galaxy.util import parse_xml_string
from galaxy.util.path import StrPath
from galaxy.util.tool_shed import common_util
from galaxy.util.tool_shed.xml_util import parse_xml

log = logging.getLogger(__name__)

DEFAULT_TOOL_SHED_URL = "https://toolshed.g2.bx.psu.edu/"
DEFAULT_TOOL_SHED_NAME = "Galaxy Main Tool Shed"
DEFAULT_TOOL_SHEDS_CONF_XML = f"""<?xml version="1.0"?>
<tool_sheds>
    <tool_shed name="{DEFAULT_TOOL_SHED_NAME}" url="{DEFAULT_TOOL_SHED_URL}"/>
</tool_sheds>
"""


class AUTH_TUPLE(NamedTuple):
    username: str
    password: str


class Registry:
    tool_sheds: Dict[str, str]
    tool_sheds_auth: Dict[str, Optional[AUTH_TUPLE]]

    def __init__(self, config: Optional[StrPath] = None):
        self.tool_sheds = {}
        self.tool_sheds_auth = {}
        if config:
            # Parse tool_sheds_conf.xml
            tree, error_message = parse_xml(config)
            if tree is None:
                log.warning(f"Unable to load references to tool sheds defined in file {str(config)}")
                return
            root = tree.getroot()
        else:
            root = parse_xml_string(DEFAULT_TOOL_SHEDS_CONF_XML)
            config = "internal default config"
        log.debug(f"Loading references to tool sheds from {config}")
        for elem in root.findall("tool_shed"):
            try:
                name = elem.get("name", None)
                url = elem.get("url", None)
                username = elem.get("user", None)
                password = elem.get("pass", None)
                if name and url:
                    self.tool_sheds[name] = url
                    self.tool_sheds_auth[name] = None
                    log.debug(f"Loaded reference to tool shed: {name}")
                if name and url and username and password:
                    self.tool_sheds_auth[name] = AUTH_TUPLE(username, password)
            except Exception as e:
                log.warning(f'Error loading reference to tool shed "{name}", problem: {str(e)}')

    def url_auth(self, url: str) -> Optional[AUTH_TUPLE]:
        """
        If the tool shed is using external auth, the client to the tool shed must authenticate to that
        as well.  This provides access to the six.moves.urllib.request.HTTPPasswordMgrWithdefaultRealm() object for the
        url passed in.

        Following more what galaxy.demo_sequencer.controllers.common does might be more appropriate at
        some stage...
        """
        shed_name = self._shed_name_for_url(url)
        if shed_name is not None:
            return self.tool_sheds_auth[shed_name]
        else:
            log.debug(f"Invalid url '{str(url)}' received by tool shed registry's url_auth method.")
            return None

    def _shed_name_for_url(self, url: str) -> Optional[str]:
        url_sans_protocol = common_util.remove_protocol_from_tool_shed_url(url)
        for shed_name, shed_url in self.tool_sheds.items():
            shed_url_sans_protocol = common_util.remove_protocol_from_tool_shed_url(shed_url)
            if url_sans_protocol.startswith(shed_url_sans_protocol):
                return shed_name
        return None

    def get_tool_shed_url(self, tool_shed: str) -> Optional[str]:
        """
        The value of tool_shed is something like: toolshed.g2.bx.psu.edu.  We need the URL to this tool shed, which is
        something like: http://toolshed.g2.bx.psu.edu/
        """
        cleaned_tool_shed = common_util.remove_protocol_from_tool_shed_url(tool_shed)
        for shed_url in self.tool_sheds.values():
            if shed_url.find(cleaned_tool_shed) >= 0:
                if shed_url.endswith("/"):
                    shed_url = shed_url.rstrip("/")
                return shed_url
        # The tool shed from which the repository was originally installed must no longer be configured in tool_sheds_conf.xml.
        return None



#!/usr/bin/env python

# Copyright (c) 2006 Bermi Ferrer Martinez
#
# bermi a-t bermilabs - com
# See the end of this file for the free software, open source license (BSD-style).
#
# Modified by the Galaxy team.

import re


class Inflector:
    """
    Inflector for pluralizing and singularizing English nouns.
    """

    # This is a small subset of words that either have the same singular and plural form, or have no singular form
    NONCHANGING_WORDS = {
        "equipment",
        "information",
        "rice",
        "money",
        "species",
        "series",
        "sheep",
        "sms",
    }

    IRREGULAR_WORDS = {
        "person": "people",
        "man": "men",
        "child": "children",
        "sex": "sexes",
        "move": "moves",
        "octopus": "octopi",
    }

    PLURALIZE_RULES = (
        ("(?i)(quiz)$", "\\1zes"),
        ("(?i)^(ox)$", "\\1en"),
        ("(?i)([m|l])ouse$", "\\1ice"),
        ("(?i)(matr|vert|ind)ix|ex$", "\\1ices"),
        ("(?i)(x|ch|ss|sh)$", "\\1es"),
        ("(?i)([^aeiouy]|qu)ies$", "\\1y"),
        ("(?i)([^aeiouy]|qu)y$", "\\1ies"),
        ("(?i)(hive)$", "\\1s"),
        ("(?i)(?:([^f])fe|([lr])f)$", "\\1\\2ves"),
        ("(?i)sis$", "ses"),
        ("(?i)([ti])um$", "\\1a"),
        ("(?i)(buffal|tomat)o$", "\\1oes"),
        ("(?i)(bu)s$", "\\1ses"),
        ("(?i)(alias|status|virus)", "\\1es"),
        ("(?i)(ax|test)is$", "\\1es"),
        ("(?i)s$", "s"),
        ("(?i)$", "s"),
    )

    SINGULARIZE_RULES = (
        ("(?i)(quiz)zes$", "\\1"),
        ("(?i)(matr)ices$", "\\1ix"),
        ("(?i)(vert|ind)ices$", "\\1ex"),
        ("(?i)^(ox)en", "\\1"),
        ("(?i)(alias|status|virus)es$", "\\1"),
        ("(?i)(cris|ax|test)es$", "\\1is"),
        ("(?i)(shoe)s$", "\\1"),
        ("(?i)(o)es$", "\\1"),
        ("(?i)(bus)es$", "\\1"),
        ("(?i)([m|l])ice$", "\\1ouse"),
        ("(?i)(x|ch|ss|sh)es$", "\\1"),
        ("(?i)(m)ovies$", "\\1ovie"),
        ("(?i)(s)eries$", "\\1eries"),
        ("(?i)([^aeiouy]|qu)ies$", "\\1y"),
        ("(?i)([lr])ves$", "\\1f"),
        ("(?i)(tive)s$", "\\1"),
        ("(?i)(hive)s$", "\\1"),
        ("(?i)([^f])ves$", "\\1fe"),
        ("(?i)(^analy)ses$", "\\1sis"),
        ("(?i)((a)naly|(b)a|(d)iagno|(p)arenthe|(p)rogno|(s)ynop|(t)he)ses$", "\\1\\2sis"),
        ("(?i)([ti])a$", "\\1um"),
        ("(?i)(n)ews$", "\\1ews"),
        ("(?i)s$", ""),
    )

    def pluralize(self, word):
        """Pluralizes nouns."""
        return self._transform(self.PLURALIZE_RULES, word)

    def singularize(self, word):
        """Singularizes nouns."""
        return self._transform(self.SINGULARIZE_RULES, word, pluralize=False)

    def cond_plural(self, number_of_records, word):
        """Returns the plural form of a word if first parameter is greater than 1"""
        if number_of_records != 1:
            return self.pluralize(word)
        return word

    def _transform(self, rules, word, pluralize=True):
        return (
            self._handle_nonchanging(word)
            or self._handle_irregular(word, pluralize=pluralize)
            or self._apply_rules(rules, word)
            or word
        )

    def _handle_nonchanging(self, word):
        lower_cased_word = word.lower()
        # Check if word is an item or the suffix of any item in NONCHANGING_WORDS
        for nonchanging_word in self.NONCHANGING_WORDS:
            if lower_cased_word.endswith(nonchanging_word):
                return word

    def _handle_irregular(self, word, pluralize=True):
        for form_a, form_b in self.IRREGULAR_WORDS.items():
            if not pluralize:
                form_a, form_b = form_b, form_a
            match = re.search(f"({form_a})$", word, re.IGNORECASE)
            if match:
                return re.sub(f"(?i){form_a}$", match.expand("\\1")[0] + form_b[1:], word)

    def _apply_rules(self, rules, word):
        for pattern, replacement in rules:
            if re.search(pattern, word):
                return re.sub(pattern, replacement, word)


# Copyright (c) 2006 Bermi Ferrer Martinez
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software to deal in this software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of this software, and to permit
# persons to whom this software is furnished to do so, subject to the following
# condition:
#
# THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THIS SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THIS SOFTWARE.



"""
Topological sort.

From Tim Peters, see:
   http://mail.python.org/pipermail/python-list/1999-July/006660.html

topsort takes a list of pairs, where each pair (x, y) is taken to
mean that x <= y wrt some abstract partial ordering.  The return
value is a list, representing a total ordering that respects all
the input constraints.
E.g.,

   topsort( [(1,2), (3,3)] )

Valid topological sorts would be any of (but nothing other than)

   [3, 1, 2]
   [1, 3, 2]
   [1, 2, 3]

... however this variant ensures that 'key' order (first element of
tuple) is preserved so the following will be result returned:

   [1, 3, 2]

because those are the permutations of the input elements that
respect the "1 precedes 2" and "3 precedes 3" input constraints.
Note that a constraint of the form (x, x) is really just a trick
to make sure x appears *somewhere* in the output list.

If there's a cycle in the constraints, say

   topsort( [(1,2), (2,1)] )

then CycleError is raised, and the exception object supports
many methods to help analyze and break the cycles.  This requires
a good deal more code than topsort itself!
"""


class CycleError(Exception):
    def __init__(self, sofar, numpreds, succs):
        Exception.__init__(self, "cycle in constraints", sofar, numpreds, succs)
        self.preds = None

    # return as much of the total ordering as topsort was able to
    # find before it hit a cycle
    def get_partial(self):
        return self[1]

    # return remaining elt -> count of predecessors map
    def get_pred_counts(self):
        return self[2]

    # return remaining elt -> list of successors map
    def get_succs(self):
        return self[3]

    # return remaining elements (== those that don't appear in
    # get_partial())
    def get_elements(self):
        return self.get_pred_counts().keys()

    # Return a list of pairs representing the full state of what's
    # remaining (if you pass this list back to topsort, it will raise
    # CycleError again, and if you invoke get_pairlist on *that*
    # exception object, the result will be isomorphic to *this*
    # invocation of get_pairlist).
    # The idea is that you can use pick_a_cycle to find a cycle,
    # through some means or another pick an (x,y) pair in the cycle
    # you no longer want to respect, then remove that pair from the
    # output of get_pairlist and try topsort again.
    def get_pairlist(self):
        succs = self.get_succs()
        answer = []
        for x in self.get_elements():
            if x in succs:
                for y in succs[x]:
                    answer.append((x, y))
            else:
                # make sure x appears in topsort's output!
                answer.append((x, x))
        return answer

    # return remaining elt -> list of predecessors map
    def get_preds(self):
        if self.preds is not None:
            return self.preds
        self.preds = preds = {}
        remaining_elts = self.get_elements()
        for x in remaining_elts:
            preds[x] = []
        succs = self.get_succs()

        for x in remaining_elts:
            if x in succs:
                for y in succs[x]:
                    preds[y].append(x)

        if __debug__:
            for x in remaining_elts:
                assert len(preds[x]) > 0
        return preds

    # return a cycle [x, ..., x] at random
    def pick_a_cycle(self):
        remaining_elts = self.get_elements()

        # We know that everything in remaining_elts has a predecessor,
        # but don't know that everything in it has a successor.  So
        # crawling forward over succs may hit a dead end.  Instead we
        # crawl backward over the preds until we hit a duplicate, then
        # reverse the path.
        preds = self.get_preds()
        from random import choice

        x = choice(remaining_elts)
        answer = []
        index = {}
        in_answer = index.has_key
        while not in_answer(x):
            index[x] = len(answer)  # index of x in answer
            answer.append(x)
            x = choice(preds[x])
        answer.append(x)
        answer = answer[index[x] :]
        answer.reverse()
        return answer


def _numpreds_and_successors_from_pairlist(pairlist):
    numpreds = {}  # elt -> # of predecessors
    successors = {}  # elt -> list of successors
    for first, second in pairlist:
        # make sure every elt is a key in numpreds
        if first not in numpreds:
            numpreds[first] = 0
        if second not in numpreds:
            numpreds[second] = 0

        # if they're the same, there's no real dependence
        if first == second:
            continue

        # since first < second, second gains a pred ...
        numpreds[second] = numpreds[second] + 1

        # ... and first gains a succ
        if first in successors:
            successors[first].append(second)
        else:
            successors[first] = [second]
    return numpreds, successors


def topsort(pairlist):
    numpreds, successors = _numpreds_and_successors_from_pairlist(pairlist)

    # suck up everything without a predecessor
    answer = [x for x in numpreds.keys() if numpreds[x] == 0]

    # for everything in answer, knock down the pred count on
    # its successors; note that answer grows *in* the loop
    for x in answer:
        assert numpreds[x] == 0
        del numpreds[x]
        if x in successors:
            for y in successors[x]:
                numpreds[y] = numpreds[y] - 1
                if numpreds[y] == 0:
                    answer.append(y)
            # following "del" isn't needed; just makes
            # CycleError details easier to grasp
            del successors[x]

    if numpreds:
        # everything in numpreds has at least one predecessor ->
        # there's a cycle
        if __debug__:
            for x in numpreds.keys():
                assert numpreds[x] > 0
        raise CycleError(answer, numpreds, successors)
    return answer


def topsort_levels(pairlist):
    numpreds, successors = _numpreds_and_successors_from_pairlist(pairlist)

    answer = []

    while 1:
        # Suck up everything without a predecessor.
        levparents = [x for x in numpreds.keys() if numpreds[x] == 0]
        if not levparents:
            break
        answer.append(levparents)
        for levparent in levparents:
            del numpreds[levparent]
            if levparent in successors:
                for levparentsucc in successors[levparent]:
                    numpreds[levparentsucc] -= 1
                del successors[levparent]

    if numpreds:
        # Everything in num_parents has at least one child ->
        # there's a cycle.
        raise CycleError(answer, numpreds, successors)

    return answer



import ipaddress
from typing import (
    List,
    Union,
)

from galaxy.util import unicodify

IpAddressT = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
IpNetworkT = Union[ipaddress.IPv4Network, ipaddress.IPv6Network]
IpAllowedListEntryT = Union[IpAddressT, IpNetworkT]


def parse_allowlist_ips(fetch_url_allowlist: List[str]) -> List[IpAllowedListEntryT]:
    return [
        ipaddress.ip_network(unicodify(ip.strip()))  # If it has a slash, assume 127.0.0.1/24 notation
        if "/" in ip
        else ipaddress.ip_address(unicodify(ip.strip()))  # Otherwise interpret it as an ip address.
        for ip in fetch_url_allowlist
        if len(ip.strip()) > 0
    ]
